"""
This module provides classes and methods to launch the LabstorIpcTest application.
LabstorIpcTest is ....
"""
from jarvis_cd.basic.pkg import Application
from jarvis_util import *


class HshmUnitTests(Application):
    """
    This class provides methods to launch the LabstorIpcTest application.
    """
    def _init(self):
        """
        Initialize paths
        """
        pass

    def _configure_menu(self):
        """
        Create a CLI menu for the configurator method.
        For thorough documentation of these parameters, view:
        https://github.com/scs-lab/jarvis-util/wiki/3.-Argument-Parsing

        :return: List(dict)
        """
        return [
            {
                'name': 'TEST_CASE',
                'msg': 'The number of threads to spawn',
                'type': int,
                'default': 1,
            },
        ]

    def _configure(self, **kwargs):
        """
        Converts the Jarvis configuration to application-specific configuration.
        E.g., OrangeFS produces an orangefs.xml file.

        :param kwargs: Configuration parameters for this pkg.
        :return: None
        """
        pass

    def start(self):
        """
        Launch an application. E.g., OrangeFS will launch the servers, clients,
        and metadata services on all necessary pkgs.

        :return: None
        """
        cmd = [
            'test_memory_cuda_mpi_exec'
        ]
        cmd = ' '.join(cmd)
        Exec(cmd,
             MpiExecInfo(env=self.env,
                         nprocs=2,
                         ppn=2,
                         do_dbg=self.config['do_dbg'],
                         dbg_port=self.config['dbg_port']))

    def stop(self):
        """
        Stop a running application. E.g., OrangeFS will terminate the servers,
        clients, and metadata services.

        :return: None
        """
        pass

    def clean(self):
        """
        Destroy all data for an application. E.g., OrangeFS will delete all
        metadata and data directories in addition to the orangefs.xml file.

        :return: None
        """
        pass



# Ported from pybind11/tests/test_chrono.py

import test_chrono_ext as m

import time
import datetime
import sys

import pytest


def test_chrono_system_clock():
    # Get the time from both c++ and datetime
    date0 = datetime.datetime.today()
    date1 = m.test_chrono1()
    date2 = datetime.datetime.today()

    # The returned value should be a datetime
    assert isinstance(date1, datetime.datetime)

    # The numbers should vary by a very small amount (time it took to execute)
    diff_python = abs(date2 - date0)
    diff = abs(date1 - date2)

    # There should never be a days difference
    assert diff.days == 0

    # Since datetime.datetime.today() calls time.time(), and on some platforms
    # that has 1 second accuracy, we compare this way
    assert diff.seconds <= diff_python.seconds


def test_chrono_system_clock_roundtrip():
    date1 = datetime.datetime.today()

    # Roundtrip the time
    date2 = m.test_chrono2(date1)

    # The returned value should be a datetime
    assert isinstance(date2, datetime.datetime)

    # They should be identical (no information lost on roundtrip)
    diff = abs(date1 - date2)
    assert diff == datetime.timedelta(0)


def test_chrono_system_clock_roundtrip_date():
    date1 = datetime.date.today()

    # Roundtrip the time
    datetime2 = m.test_chrono2(date1)
    date2 = datetime2.date()
    time2 = datetime2.time()

    # The returned value should be a datetime
    assert isinstance(datetime2, datetime.datetime)
    assert isinstance(date2, datetime.date)
    assert isinstance(time2, datetime.time)

    # They should be identical (no information lost on roundtrip)
    diff = abs(date1 - date2)
    assert diff.days == 0
    assert diff.seconds == 0
    assert diff.microseconds == 0

    # Year, Month & Day should be the same after the round trip
    assert date1 == date2

    # There should be no time information
    assert time2.hour == 0
    assert time2.minute == 0
    assert time2.second == 0
    assert time2.microsecond == 0


SKIP_TZ_ENV_ON_WIN = pytest.mark.skipif(
    "sys.platform == 'win32'",
    reason="TZ environment variable only supported on POSIX"
)


@pytest.mark.parametrize(
    "time1",
    [
        datetime.datetime.today().time(),
        datetime.time(0, 0, 0),
        datetime.time(0, 0, 0, 1),
        datetime.time(0, 28, 45, 109827),
        datetime.time(0, 59, 59, 999999),
        datetime.time(1, 0, 0),
        datetime.time(5, 59, 59, 0),
        datetime.time(5, 59, 59, 1),
    ],
)
@pytest.mark.parametrize(
    "tz",
    [
        None,
        pytest.param("Europe/Brussels", marks=SKIP_TZ_ENV_ON_WIN),
        pytest.param("Asia/Pyongyang", marks=SKIP_TZ_ENV_ON_WIN),
        pytest.param("America/New_York", marks=SKIP_TZ_ENV_ON_WIN),
    ],
)
def test_chrono_system_clock_roundtrip_time(time1, tz, monkeypatch):
    if tz is not None:
        monkeypatch.setenv("TZ", f"/usr/share/zoneinfo/{tz}")

    # Roundtrip the time
    datetime2 = m.test_chrono2(time1)
    date2 = datetime2.date()
    time2 = datetime2.time()

    # The returned value should be a datetime
    assert isinstance(datetime2, datetime.datetime)
    assert isinstance(date2, datetime.date)
    assert isinstance(time2, datetime.time)

    # Hour, Minute, Second & Microsecond should be the same after the round trip
    assert time1 == time2

    # There should be no date information (i.e. date = python base date)
    assert date2.year == 1970
    assert date2.month == 1
    assert date2.day == 1


def test_chrono_duration_roundtrip():
    # Get the difference between two times (a timedelta)
    date1 = datetime.datetime.today()
    date2 = datetime.datetime.today()
    diff = date2 - date1

    # Make sure this is a timedelta
    assert isinstance(diff, datetime.timedelta)

    cpp_diff = m.test_chrono3(diff)

    assert cpp_diff == diff

    # Negative timedelta roundtrip
    diff = datetime.timedelta(microseconds=-1)
    cpp_diff = m.test_chrono3(diff)

    assert cpp_diff == diff


def test_chrono_duration_subtraction_equivalence():
    date1 = datetime.datetime.today()
    date2 = datetime.datetime.today()

    diff = date2 - date1
    cpp_diff = m.test_chrono4(date2, date1)

    assert cpp_diff == diff


def test_chrono_duration_subtraction_equivalence_date():
    date1 = datetime.date.today()
    date2 = datetime.date.today()

    diff = date2 - date1
    cpp_diff = m.test_chrono4(date2, date1)

    assert cpp_diff == diff


def test_chrono_steady_clock():
    time1 = m.test_chrono5()
    assert isinstance(time1, datetime.timedelta)


def test_chrono_steady_clock_roundtrip():
    time1 = datetime.timedelta(days=10, seconds=10, microseconds=100)
    time2 = m.test_chrono6(time1)

    assert isinstance(time2, datetime.timedelta)

    # They should be identical (no information lost on roundtrip)
    assert time1 == time2

    # Floating point conversion also works
    assert m.test_chrono6(time1.total_seconds()) == time1


def test_floating_point_duration():
    # Test using a floating point number in seconds
    time = m.test_chrono7(35.525123)

    assert isinstance(time, datetime.timedelta)

    assert time.seconds == 35
    assert 525122 <= time.microseconds <= 525123

    diff = m.test_chrono_float_diff(43.789012, 1.123456)
    assert diff.seconds == 42
    assert 665556 <= diff.microseconds <= 665557


def test_nano_timepoint():
    time = datetime.datetime.now()
    time1 = m.test_nano_timepoint(time, datetime.timedelta(seconds=60))
    assert time1 == time + datetime.timedelta(seconds=60)


def test_chrono_different_resolutions():
    resolutions = m.different_resolutions()
    time = datetime.datetime.now()
    resolutions.timestamp_h = time
    resolutions.timestamp_m = time
    resolutions.timestamp_s = time
    resolutions.timestamp_ms = time
    resolutions.timestamp_us = time

    assert time == resolutions.timestamp_us
    time = time.replace(microsecond=(time.microsecond // 1000) * 1000)
    assert time == resolutions.timestamp_ms
    time = time.replace(microsecond=0)
    assert time == resolutions.timestamp_s
    time = time.replace(second=0)
    assert time == resolutions.timestamp_m
    time = time.replace(minute=0)
    assert time == resolutions.timestamp_h


# Tests below this point are new in nanobind


def test_chrono_misc():
    from datetime import datetime, timedelta

    advance_datetime = m.test_nano_timepoint
    difference_between_datetimes = m.test_nano_timepoint_diff
    roundtrip_datetime = m.test_nano_timepoint_roundtrip

    d1 = datetime(2023, 4, 5, 12, 0, 0, 0)
    d2 = datetime(2023, 4, 5, 12, 30, 0, 123)
    # datetime -> time_point and duration -> timedelta conversion
    assert difference_between_datetimes(d1, d2) == d1 - d2
    assert difference_between_datetimes(d2, d1) == d2 - d1
    # date -> time_point conversion
    assert difference_between_datetimes(d2, d1.date()) == timedelta(
        hours=12, minutes=30, microseconds=123
    )
    # time -> time_point conversion
    assert difference_between_datetimes(d2.time(), d1.time()) == timedelta(
        minutes=30, microseconds=123
    )
    assert roundtrip_datetime(d1.time()) == datetime(1970, 1, 1, 12, 0, 0)
    for td in (
        timedelta(seconds=5),
        timedelta(microseconds=123),
        timedelta(days=1, seconds=10),
        timedelta(seconds=-5),
        timedelta(microseconds=-123),
        timedelta(days=-1, seconds=-10),
    ):
        # timedelta -> duration conversion
        assert advance_datetime(d1, td) == d1 + td
        # float -> duration conversion
        assert advance_datetime(d1, td.total_seconds()) == d1 + td
    # time_point -> datetime conversion
    assert roundtrip_datetime(d1) == d1
    assert roundtrip_datetime(d2) == d2


@pytest.mark.parametrize(
    "test_type,roundtrip_name",
    [
        (datetime.timedelta, 'test_chrono7'),
        (datetime.datetime, 'test_nano_timepoint_roundtrip'),
    ]
)
def test_chrono_invalid(test_type, roundtrip_name):
    roundtrip = getattr(m, roundtrip_name)

    # Can't pass None or an integer where a duration or timepoint is expected
    with pytest.raises(TypeError, match="incompatible function arguments"):
        roundtrip(None)
    with pytest.raises(TypeError, match="incompatible function arguments"):
        roundtrip(42)

    # Can't pass a duration where a timepoint is expected, or vice versa
    with pytest.raises(TypeError, match="incompatible function arguments"):
        if test_type is datetime.datetime:
            roundtrip(datetime.timedelta(seconds=5))
        else:
            roundtrip(datetime.datetime.now())

    # On the limited API we access timedelta/datetime objects via
    # regular attribute access; test that invalid results are handled
    # reasonably. On the full API we use Python's <datetime.h> header
    # so we'll always access the true C-level datetime slot and can't
    # be fooled by tricks like this. PyPy uses normal attribute access
    # and works like the limited API in this respect.

    class fake_type(test_type):
        @property
        def seconds(self):
            return self.override_value

        @property
        def second(self):
            return self.override_value

    if test_type is datetime.datetime:
        fake_val = fake_type.fromtimestamp(time.time())
        replace_overridden = lambda s: fake_val.replace(second=s)
    else:
        fake_val = fake_type(days=1, seconds=10, microseconds=123456)
        replace_overridden = lambda s: fake_type(
            days=1, seconds=s, microseconds=123456
        )

    for fake_result, errtype in (
        ("hi", "TypeError"),
        (0, None),
        (2**64, "Python int too large to convert to C long"),
        (2**32, "OverflowError"),
    ):
        fake_val.override_value = fake_result
        if not m.access_via_python:
            assert roundtrip(fake_val) == fake_val
        elif errtype is None:
            assert roundtrip(fake_val) == replace_overridden(fake_result)
        elif test_type is datetime.timedelta and sys.implementation.name == "pypy":
            # pypy's cpyext module converts timedelta to a C structure
            # before the nanobind function even gets called, producing
            # a different exception than the one we're testing below.
            # datetime still works as it doesn't have its attributes
            # converted but instead is implemented with Python
            # attribute accesses.
            pass
        else:
            from test.support import catch_unraisable_exception

            with catch_unraisable_exception() as cm:
                with pytest.raises(TypeError, match="incompatible function arguments"):
                    roundtrip(fake_val)
                assert cm.unraisable is not None
                assert errtype in repr(cm.unraisable.exc_value)



import asyncio
from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client

async def main():
    # Use stdio_client to launch the server as a subprocess for testing
    server_params = StdioServerParameters(
        command="python",
        args=["server/server.py"],
        env=None
    )
    async with stdio_client(server_params) as (read, write):
        async with ClientSession(read, write) as session:
            await session.initialize()

            print("Available tools:")
            tools = await session.list_tools()
            for tool in tools:
                print("-", tool)

            print("\nAvailable resources:")
            resources = await session.list_resources()
            for resource in resources:
                print("-", resource)

            # List all HDF5/HDF4 paths using the tool
            result = await session.call_tool("list_hdf_paths")
            print("\nHDF5/HDF4 paths in server:")
            print(result)

            # Fetch a specific HDF5 dataset as a resource
            hdf5_dataset_path = "group1/dataset1"
            hdf5_resource_uri = f"hdf://hdf5/{hdf5_dataset_path}"
            content = await session.read_resource(hdf5_resource_uri)
            print(f"\nFetched HDF5 resource '{hdf5_resource_uri}':")
            print(content)

            # Fetch a specific HDF4 dataset as a resource
            hdf4_dataset_path = "dataset1"
            hdf4_resource_uri = f"hdf://hdf4/{hdf4_dataset_path}"
            content = await session.read_resource(hdf4_resource_uri)
            print(f"\nFetched HDF4 resource '{hdf4_resource_uri}':")
            print(content)

if __name__ == "__main__":
    asyncio.run(main())



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

""" Testing model for array types. """

from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
import os.path as op

import logging
log = logging.getLogger(__name__)

from hdf_compass import compass_model

DT_CMP = np.dtype([('a', 'i'), ('b', 'f')])

DATA = {'array://localhost/a_0d': np.array(1),
        'array://localhost/a_1d': np.arange(10),
        'array://localhost/a_2d': np.arange(10 * 10).reshape((10, 10)),
        'array://localhost/a_3d': np.arange(10 * 10 * 10).reshape((10, 10, 10)),
        'array://localhost/a_4d': np.arange(10 * 10 * 10 * 10).reshape((10, 10, 10, 10)),
        'array://localhost/cmp_0d': np.array((1, 2), dtype=DT_CMP),
        'array://localhost/cmp_1d': np.ones((10,), dtype=DT_CMP),
        'array://localhost/cmp_2d': np.ones((10, 10), dtype=DT_CMP),
        'array://localhost/cmp_3d': np.ones((10, 10, 10), dtype=DT_CMP),
        'array://localhost/cmp_4d': np.ones((10, 10, 10, 10), dtype=DT_CMP),
        'array://localhost/S_0d': np.array(b"Hello"),
        'array://localhost/S_1d': np.array([b"Hello", b"Ciao"]),
        'array://localhost/S_2d': np.array([[b"Hello", b"Ciao"], [b"Hello", b"Ciao"]]),
        'array://localhost/S_3d': np.array([[[b"Hello", b"Ciao"], [b"Hello", b"Ciao"]],
                                            [[b"Hello", b"Ciao"], [b"Hello", b"Ciao"]]]),
        'array://localhost/U_1d': np.array(["Hello", "Ciao"]),
        'array://localhost/U_2d': np.array([["Hello", "Ciao"], ["Hello", "Ciao"]]),
        'array://localhost/U_3d': np.array([[["Hello", "Ciao"], ["Hello", "Ciao"]],
                                            [["Hello", "Ciao"], ["Hello", "Ciao"]]]),
        'array://localhost/v_0d': np.array(b'\x01', dtype='|V1'),
        'array://localhost/non_square': np.arange(5 * 10).reshape((5, 10)),
        }


class ArrayStore(compass_model.Store):
    """
        A "data store" represented by a set of arrays in memory.

        Keys are array names as reported in DATA.
    """

    @staticmethod
    def plugin_name():
        return "Array"

    @staticmethod
    def plugin_description():
        return "A testing plugin used to evaluate the array visualization."

    def __contains__(self, key):
        if (key == '/') or (key is None):
            log.debug("is root: %s" % key)
            return True
        return key in DATA

    @property
    def url(self):
        return self._url

    @property
    def display_name(self):
        return "Testing arrays"

    @property
    def root(self):
        return self['/']

    @property
    def valid(self):
        return self._valid

    @staticmethod
    def can_handle(url):
        if url == "array://localhost":
            log.debug("able to handle %s? yes" % url)
            return True
        log.debug("able to handle %s? no" % url)
        return False

    def __init__(self, url):
        if not self.can_handle(url):
            raise ValueError(url)
        self._url = url
        self._valid = True

    def close(self):
        self._valid = False

    def get_parent(self, key):
        if key is None:
            return None
        if key == "/":
            return None
        return self[op.dirname(key)]


class ArrayContainer(compass_model.Container):
    """ Represents a container of in-memory array. """

    class_kind = "Array Container"

    @staticmethod
    def can_handle(store, key):
        return (key == '/') or (key is None)

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self._names = sorted(DATA.keys())

    @property
    def key(self):
        return None

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return "/"

    @property
    def description(self):
        return self.display_name

    def __len__(self):
        return len(DATA)

    def __iter__(self):
        for name in self._names:
            yield self._store[name]

    def __getitem__(self, idx):
        key = self._names[idx]
        return self._store[key]


class Array(compass_model.Array):
    """ An N-D array """

    class_kind = "TestArray"

    @staticmethod
    def can_handle(store, key):
        return key in DATA

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self.data = DATA[key]

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return self.key.rsplit('/', 1)[-1]

    @property
    def description(self):
        return self.display_name

    @property
    def shape(self):
        return self.data.shape

    @property
    def dtype(self):
        return self.data.dtype

    def __getitem__(self, args):
        return self.data[args]

    def is_plottable(self):
        if self.dtype.kind == 'S':
            log.debug("Not plottable since ASCII String (characters: %d)" % self.dtype.itemsize)
            return False
        if self.dtype.kind == 'U':
            log.debug("Not plottable since Unicode String (characters: %d)" % self.dtype.itemsize)
            return False
        return True


class ArrayText(compass_model.Text):
    """ Represents a text array (both ASCII and UNICODE). """

    class_kind = "TestArray [text]"

    @staticmethod
    def can_handle(store, key):
        if key not in DATA:
            return False

        if DATA[key].dtype.kind == 'S':
            # log.debug("ASCII String (characters: %d)" % DATA[key].dtype.itemsize)
            return True
        if DATA[key].dtype.kind == 'U':
            # log.debug("Unicode String (characters: %d)" % DATA[key].dtype.itemsize)
            return True
        return False

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self.data = DATA[key]

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return self.key.rsplit('/', 1)[-1]

    @property
    def description(self):
        return 'Text "%s"' % (self.display_name,)

    @property
    def shape(self):
        return self.data.shape

    @property
    def text(self):
        txt = str()

        if len(self.shape) == 0:
            print(self.data)
            txt += str(self.data)

        elif len(self.shape) == 1:
            for el in self.data:
                txt += el + ",\n"

        elif len(self.shape) == 2:
            for i in range(self.shape[0]):
                for j in range(self.shape[1]):
                    txt += self.data[i, j] + ", "
                txt += "\n"

        else:
            txt = ">> display of more than 2D string array not implemented <<"

        return txt


class ArrayKV(compass_model.KeyValue):
    class_kind = "Array Key/Value Attrs"

    @staticmethod
    def can_handle(store, key):
        return key in DATA

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self.data = {'a': np.array((1, 2)), 'b': np.array("Hello"), 'c': np.array(b'\x01', dtype='|V1')}

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return self.key.rsplit('/', 1)[-1]

    @property
    def description(self):
        return self.display_name

    @property
    def keys(self):
        return self.data.keys()

    def __getitem__(self, args):
        return self.data[args]


ArrayStore.push(ArrayKV)
ArrayStore.push(ArrayContainer)
ArrayStore.push(Array)
ArrayStore.push(ArrayText)

compass_model.push(ArrayStore)



#!/usr/bin/env python3
"""
Compilation-focused test for WRP CTE Core Python bindings
Tests that all bindings compile and are accessible without full runtime initialization
"""

import sys
import os

# When running with python -I (isolated mode), we need to manually add the current directory
# The test is run with WORKING_DIRECTORY set to the module directory
sys.path.insert(0, os.getcwd())

def test_import():
    """Test that the module can be imported successfully"""
    try:
        import wrp_cte_core_ext as cte
        print("✅ Module import successful")
        return True
    except ImportError as e:
        print(f"❌ Module import failed: {e}")
        return False

def test_basic_types():
    """Test that basic types can be created"""
    try:
        import wrp_cte_core_ext as cte
        
        # Test enum
        op = cte.CteOp.kPutBlob
        print(f"✅ CteOp enum created: {op}")
        
        # Test TagId
        tag_id = cte.TagId()
        null_tag = cte.TagId.GetNull()
        print(f"✅ TagId created, null check: {null_tag.IsNull()}")
        
        # Test BlobId
        blob_id = cte.BlobId()
        null_blob = cte.BlobId.GetNull()
        print(f"✅ BlobId created, null check: {null_blob.IsNull()}")
        
        # Test MemContext
        mctx = cte.MemContext()
        print("✅ MemContext created")
        
        # Test CteTelemetry
        telemetry = cte.CteTelemetry()
        print(f"✅ CteTelemetry created with op: {telemetry.op_}")
        
        # Test Client type accessibility (without instantiation)
        client_type = cte.Client
        print(f"✅ Client type accessible: {client_type}")
        
        return True
    except Exception as e:
        print(f"❌ Basic types test failed: {e}")
        return False

def test_functions_accessibility():
    """Test that module functions are accessible without calling them"""
    try:
        import wrp_cte_core_ext as cte

        # Test initialization functions exist
        runtime_init = cte.chimaera_runtime_init
        client_init = cte.chimaera_client_init
        cte_init = cte.initialize_cte
        print(f"✅ chimaera_runtime_init function accessible: {runtime_init}")
        print(f"✅ chimaera_client_init function accessible: {client_init}")
        print(f"✅ initialize_cte function accessible: {cte_init}")

        # Test client getter function exists
        client_getter = cte.get_cte_client
        print(f"✅ get_cte_client function accessible: {client_getter}")

        return True
    except Exception as e:
        print(f"❌ Functions accessibility test failed: {e}")
        return False

def test_poll_telemetry_log_binding():
    """Test PollTelemetryLog method binding without runtime initialization"""
    try:
        import wrp_cte_core_ext as cte
        
        # Test Client type has PollTelemetryLog method
        client_type = cte.Client
        
        # Check if PollTelemetryLog method exists on the type
        if hasattr(client_type, 'PollTelemetryLog'):
            print("✅ PollTelemetryLog method exists on Client class")
        else:
            print("❌ PollTelemetryLog method not found on Client class")
            return False
        
        # Test MemContext creation for method calls
        mctx = cte.MemContext()
        print("✅ Created MemContext for potential method calls")
        
        print("✅ PollTelemetryLog binding verification successful")
        return True
    except Exception as e:
        print(f"❌ PollTelemetryLog binding test failed: {e}")
        return False

def test_runtime_functions_signature():
    """Test that runtime initialization functions have correct signatures"""
    try:
        import wrp_cte_core_ext as cte
        import inspect

        # Check chimaera_runtime_init signature
        runtime_init_sig = inspect.signature(cte.chimaera_runtime_init)
        print(f"✅ chimaera_runtime_init signature: {runtime_init_sig}")

        # Check chimaera_client_init signature
        client_init_sig = inspect.signature(cte.chimaera_client_init)
        print(f"✅ chimaera_client_init signature: {client_init_sig}")

        # Check initialize_cte signature
        cte_init_sig = inspect.signature(cte.initialize_cte)
        print(f"✅ initialize_cte signature: {cte_init_sig}")

        # Check get_cte_client signature
        client_getter_sig = inspect.signature(cte.get_cte_client)
        print(f"✅ get_cte_client signature: {client_getter_sig}")

        return True
    except Exception as e:
        print(f"❌ Function signature test failed: {e}")
        return False

def main():
    """Run all compilation-focused tests"""
    print("🧪 Running Python bindings compilation verification tests...")
    
    tests = [
        ("Module Import", test_import),
        ("Basic Types", test_basic_types),
        ("Functions Accessibility", test_functions_accessibility),
        ("PollTelemetryLog Binding", test_poll_telemetry_log_binding),
        ("Runtime Functions Signature", test_runtime_functions_signature)
    ]
    
    passed = 0
    total = len(tests)
    
    for test_name, test_func in tests:
        print(f"\n📋 Testing {test_name}...")
        if test_func():
            passed += 1
        else:
            print(f"❌ {test_name} failed")
    
    print(f"\n📊 Test Results: {passed}/{total} tests passed")
    
    if passed == total:
        print("🎉 All Python bindings compilation tests passed!")
        return 0
    else:
        print("💥 Some tests failed!")
        return 1

if __name__ == "__main__":
    sys.exit(main())


"""
Download additional data files from AWS
"""

from __future__ import absolute_import, division, print_function

import os.path
try:
    import wget

except ImportError as e:
    print("missing wget, try to install it: pip install wget: %s" % e)
    import sys

    sys.exit(1)

files = [
    'ami_hdf.h5',
    'attrfile.h5',
    'craterlake.h5',
    'compound.h5',
    'comp_complex.h5',
    'countries.h5',
    'hapmap_compressed.h5',
    'hdf5_test.h5',
    'iso++.h5',
    'sequences.h5',
    'snps1000_chr22_CEU.h5',
    'LD_22_5Kx5K_matrix.h5',  # 5 MB
    'shuttle_hdf.h5',  # 8 MB
    'full8_400um.mnc.h5',  # 10MB
    'DOQ.h5',  # 45 MB
    'Sample_Urban_Data.h5',  # 145 MB
    'h5pf_md_manygroupnew2.h5',  # 155 MB
    'gz6_SCRIS_npp_d20140522_t0754579_e0802557_b13293__noaa_pop.h5',  # 187 MB
    # 'NARA_TWR.h5',  # 3 GB!  Uncomment if you really want this.
]

for filename in files:
    uri = 'https://s3.amazonaws.com/hdfgroup/data/hdf5test/' + filename
    print( uri)
    if os.path.isfile(filename):
        print(filename, "already downloaded - skipping")
    else:
        wget.download(uri, bar=wget.bar_thermometer)
print("done!")



#!/usr/bin/env python3
"""
get_oauth_token.py - Get Globus OAuth2 tokens programmatically

This demonstrates how to get Globus access tokens using the OAuth2 flow
for REST API access to download data.

For production use, you should:
1. Register a Native App or Confidential Client at https://app.globus.org/settings/developers
2. Use the client ID and implement proper OAuth2 flow

This script uses the Globus SDK's Native App flow.
"""

import sys
import globus_sdk

def get_tokens_for_collection(collection_id):
    """
    Get OAuth2 tokens for Transfer API and HTTPS access to a collection.

    Args:
        collection_id: The Globus collection/endpoint ID

    Returns:
        Dictionary with access tokens
    """

    # Use the Globus SDK's tutorial client ID (for demonstration only)
    # For production, register your own app at https://app.globus.org/settings/developers
    CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"  # Tutorial Native App client ID

    # Define the scopes we need
    # 1. Transfer API scope - for getting endpoint details
    transfer_scope = "urn:globus:auth:scope:transfer.api.globus.org:all"

    # 2. Collection-specific HTTPS scope - for downloading files via HTTPS
    https_scope = f"https://auth.globus.org/scopes/{collection_id}/https"

    # Combine scopes
    scopes = [transfer_scope, https_scope]

    print("=== Globus OAuth2 Token Generation ===")
    print("")
    print(f"Collection ID: {collection_id}")
    print(f"Required scopes:")
    for scope in scopes:
        print(f"  - {scope}")
    print("")

    # Create a Native App client
    client = globus_sdk.NativeAppAuthClient(CLIENT_ID)

    # Start the OAuth2 flow
    client.oauth2_start_flow(requested_scopes=scopes)

    # Get the authorization URL
    authorize_url = client.oauth2_get_authorize_url()

    print("Please visit this URL to authorize the application:")
    print("")
    print(authorize_url)
    print("")

    # Wait for the user to authorize and paste the code
    auth_code = input("Paste the authorization code here: ").strip()

    # Exchange the code for tokens
    token_response = client.oauth2_exchange_code_for_tokens(auth_code)

    # Get tokens for each resource server
    transfer_tokens = token_response.by_resource_server['transfer.api.globus.org']

    # The HTTPS scope is a dependent scope, so it might be in the transfer tokens
    # or we need to get it separately

    print("")
    print("=== Tokens Retrieved ===")
    print("")
    print("Transfer API Access Token:")
    print(transfer_tokens['access_token'])
    print("")

    # Check if we have HTTPS scope tokens
    # For dependent scopes, they may be included in the transfer token response
    print("You can now use this token with:")
    print(f"  export GLOBUS_ACCESS_TOKEN='{transfer_tokens['access_token']}'")
    print("")

    # Save tokens to a file
    with open('/tmp/globus_tokens.txt', 'w') as f:
        f.write(f"GLOBUS_ACCESS_TOKEN={transfer_tokens['access_token']}\n")

    print("Token saved to: /tmp/globus_tokens.txt")
    print("Load it with: source <(cat /tmp/globus_tokens.txt | sed 's/^/export /')")

    return {
        'transfer': transfer_tokens['access_token'],
        'refresh': transfer_tokens.get('refresh_token')
    }

if __name__ == "__main__":
    collection_id = "e8cf0e9a-f96a-11ed-9a83-83ef71fbf0ae"

    if len(sys.argv) > 1:
        collection_id = sys.argv[1]

    try:
        tokens = get_tokens_for_collection(collection_id)
    except Exception as e:
        print(f"ERROR: {e}", file=sys.stderr)
        sys.exit(1)



from hdf_compass.compass_model.test import container, store
from hdf_compass.opendap_model import Server, Dataset

url = "http://test.opendap.org/opendap/hyrax/data/hdf5/grid_1_2d.h5"
s_1 = store(Server, url)

url = "http://test.opendap.org/opendap/hyrax/data/nc/bears.nc"
s_2 = store(Server, url)


from jarvis_cd.basic.pkg import Application
from jarvis_util.shell.exec import Exec
from jarvis_util.shell.pssh_exec import PsshExecInfo
from jarvis_util import *
import os

class LightbeamDistributedTest(Application):
    def _init(self):
        pass

    def _configure_menu(self):
        return [
            {'name': 'hostfile', 'msg': 'Path to hostfile', 'type': str, 'default': None},
            {'name': 'transport', 'msg': 'Transport (zeromq or thallium)', 'type': str, 'default': 'thallium'},
            {'name': 'protocol', 'msg': 'Protocol (tcp or ofi+sockets)', 'type': str, 'default': 'ofi+sockets'},
            {'name': 'domain', 'msg': 'Domain (e.g., localhost)', 'type': str, 'default': 'localhost'},
            {'name': 'port', 'msg': 'Port (e.g., 8200)', 'type': int, 'default': 8200},
            {'name': 'num_msgs', 'msg': 'Number of messages per node', 'type': int, 'default': 10},
            {'name': 'msg_size', 'msg': 'Size of each message', 'type': int, 'default': 32},
        ]

    def _configure(self, **kwargs):
        self.update_config(kwargs, rebuild=False)
        self.hostfile = self.config['hostfile']
        self.transport = self.config['transport']
        self.protocol = self.config['protocol']
        self.domain = self.config['domain']
        self.port = self.config['port']

    def start(self):
        self.binary = 'distributed_lightbeam_test'
        # Replace None or empty string with '' for all parameters before use
        def quote_if_empty(val):
            return "''" if val is None or (isinstance(val, str) and val.strip() == '') else str(val)
        self.transport = quote_if_empty(self.config['transport'])
        self.hostfile = quote_if_empty(self.config['hostfile'])
        self.protocol = quote_if_empty(self.config['protocol'])
        self.domain = quote_if_empty(self.config['domain'])
        self.port = quote_if_empty(self.config['port'])
        cmd = f"{self.binary} {self.transport} {self.hostfile} {self.protocol} {self.domain} {self.port}"
        # Add num_msgs and msg_size if present in config
        num_msgs = self.config.get('num_msgs')
        msg_size = self.config.get('msg_size')
        if num_msgs is not None:
            cmd += f" {num_msgs}"
        if msg_size is not None:
            cmd += f" {msg_size}"
        print(f"[LightbeamDistributedTest] Launching: {cmd}")
        hosts = Hostfile(path=self.config['hostfile'])
        print(f"[LightbeamDistributedTest] Hosts: {hosts} , len(hosts): {len(hosts)}", hosts.path)
        print(f" cmd: {cmd}")
        # hosts = self.jarvis.hostfile
        Exec(cmd, MpiExecInfo(hosts= hosts, env=self.env, exec_async=False, nprocs=len(hosts), ppn=2))

    def stop(self):
        """
        Stop a running application. E.g., terminate servers, clients, etc.
        """
        pass

    def clean(self):
        """
        Destroy all data for the application. E.g., delete metadata, data directories, etc.
        """
        pass 


import test_functions_ext as t
import pytest
import sys
import re

# Some helper declaration to check types across different Python versions
if sys.version_info < (3, 9):
    TYPING_TUPLE = "typing.Tuple"
else:
    TYPING_TUPLE = "tuple"

# Reference counting behavior changed on 3.14a7+
py_3_14a7_or_newer = sys.version_info >= (3, 14, 0, 'alpha', 7)

def fail_fn():  # used in test_30
    raise RuntimeError("Foo")


def test01_capture():
    # Functions with and without capture object of different sizes
    assert t.test_01() is None
    assert t.test_02(5, 3) == 2
    assert t.test_03(5, 3) == 44
    assert t.test_04() == 60
    assert t.test_simple(0, 1, 2, 3, 4, 5, 6, 7) == 14


def test02_default_args():
    # Default arguments
    assert t.test_02() == 7
    assert t.test_02(7) == 6


def test03_kwargs():
    # Basic use of keyword arguments
    assert t.test_02(3, 5) == -2
    assert t.test_02(3, down=5) == -2
    assert t.test_02(down=5, up=3) == -2

    # Make sure non-interned keyword names work also
    i_cant_believe_its_not_down = "".join("down")
    assert i_cant_believe_its_not_down is not "down"
    assert t.test_02(**{i_cant_believe_its_not_down: 5, "up": 3}) == -2
    assert t.test_02(**{i_cant_believe_its_not_down: 5}) == 3
    with pytest.raises(TypeError):
        t.test_02(unexpected=27)
    with pytest.raises(TypeError):
        t.test_02(**{i_cant_believe_its_not_down: None})


def test04_overloads():
    assert t.test_05(0) == 1
    assert t.test_05(0.0) == 2


def test05_signature():
    assert t.test_01.__doc__ == "test_01() -> None"
    assert t.test_02.__doc__ == "test_02(up: int = 8, down: int = 1) -> int"
    assert t.test_05.__doc__ == (
        "test_05(arg: int, /) -> int\n"
        "test_05(arg: float, /) -> int\n"
        "\n"
        "Overloaded function.\n"
        "\n"
        "1. ``test_05(arg: int, /) -> int``\n"
        "\n"
        "doc_1\n"
        "\n"
        "2. ``test_05(arg: float, /) -> int``\n"
        "\n"
        "doc_2"
    )

    assert t.test_05b.__doc__ == (
        "test_05b(arg: int, /) -> int\n"
        "test_05b(arg: float, /) -> int\n"
        "\n"
        "doc_1"
    )

    assert t.test_05c.__doc__ == (
        "test_05c(arg: int, /) -> int\n"
        "test_05c(arg: float, /) -> int\n"
        "\n"
        "Overloaded function.\n"
        "\n"
        "1. ``test_05c(arg: int, /) -> int``\n"
        "\n"
        "doc_1\n"
        "\n"
        "2. ``test_05c(arg: float, /) -> int``\n"
    )

    assert t.test_07.__doc__ == (
        f"test_07(arg0: int, arg1: int, /, *args, **kwargs) -> {TYPING_TUPLE}[int, int]\n"
        f"test_07(a: int, b: int, *myargs, **mykwargs) -> {TYPING_TUPLE}[int, int]"
    )


def test06_signature_error():
    with pytest.raises(TypeError) as excinfo:
        t.test_05("x", y=4)
    assert str(excinfo.value) == (
        "test_05(): incompatible function arguments. The "
        "following argument types are supported:\n"
        "    1. test_05(arg: int, /) -> int\n"
        "    2. test_05(arg: float, /) -> int\n\n"
        "Invoked with types: str, kwargs = { y: int }"
    )


def test07_raises():
    with pytest.raises(RuntimeError) as excinfo:
        t.test_06()
    assert str(excinfo.value) == "oops!"


def test08_args_kwargs():
    assert t.test_07(1, 2) == (0, 0)
    assert t.test_07(a=1, b=2) == (0, 0)
    assert t.test_07(a=1, b=2, c=3) == (0, 1)
    assert t.test_07(1, 2, 3, c=4) == (1, 1)
    assert t.test_07(1, 2, 3, 4, c=5, d=5) == (2, 2)


def test09_maketuple():
    assert t.test_tuple() == ("Hello", 123)
    with pytest.raises(RuntimeError) as excinfo:
        assert t.test_bad_tuple()
    value = str(excinfo.value)
    assert value == "std::bad_cast" or value == "bad cast"


def test10_cpp_call_simple():
    result = []

    def my_callable(a, b):
        result.append((a, b))

    t.test_call_2(my_callable)
    assert result == [(1, 2)]

    with pytest.raises(TypeError) as excinfo:
        t.test_call_1(my_callable)
    assert "my_callable() missing 1 required positional argument: 'b'" in str(
        excinfo.value
    )
    assert result == [(1, 2)]


def test11_call_complex():
    result = []

    def my_callable(*args, **kwargs):
        result.append((args, kwargs))

    t.test_call_extra(my_callable)
    assert result == [((1, 2), {"extra": 5})]

    result.clear()
    t.test_call_extra(my_callable, 5, 6, hello="world")
    assert result == [((1, 2, 5, 6), {"extra": 5, "hello": "world"})]


def test12_list_tuple_manipulation():
    li = [1, 5, 6, 7]
    t.test_list(li)
    assert li == [1, 5, 123, 7, 19]

    tu = (1, 5, 6, 7)
    assert t.test_tuple(tu) == 19
    assert tu == (1, 5, 6, 7)


def test13_call_guard():
    assert t.call_guard_value() == 0
    assert t.test_call_guard() == 1
    assert t.call_guard_value() == 2
    assert t.test_call_guard_wrapper_rvalue_ref(1) == 1
    assert not t.test_release_gil()


def test14_print(capsys):
    t.test_print()
    captured = capsys.readouterr()
    assert captured.out == "Test 1\nTest 2\n"


def test15_iter():
    assert t.test_iter(()) == []
    assert t.test_iter((1,)) == [1]
    assert t.test_iter((1, 2)) == [1, 2]
    assert t.test_iter((1, 2, 3)) == [1, 2, 3]


def test16_iter_tuple():
    assert t.test_iter_tuple(()) == []
    assert t.test_iter_tuple((1,)) == [1]
    assert t.test_iter_tuple((1, 2)) == [1, 2]
    assert t.test_iter_tuple((1, 2, 3)) == [1, 2, 3]


def test17_iter_tuple():
    assert t.test_iter_list([]) == []
    assert t.test_iter_list([1]) == [1]
    assert t.test_iter_list([1, 2]) == [1, 2]
    assert t.test_iter_list([1, 2, 3]) == [1, 2, 3]


def test18_raw_doc():
    assert (
        t.test_08.__doc__
        == """test_08(arg: int, /) -> int
test_08(x: typing.Annotated[float, 'foo']) -> int

Overloaded function.

1. ``test_08(arg: int, /) -> int``

first docstring

2. ``test_08(x: typing.Annotated[float, 'foo']) -> int``

another docstring"""
    )


def test19_type_check_manual():
    assert t.test_09.__doc__ == "test_09(arg: type, /) -> bool"

    assert t.test_09(bool) is True
    assert t.test_09(int) is False
    with pytest.raises(TypeError) as excinfo:
        assert t.test_09(True)
    assert "incompatible function arguments" in str(excinfo.value)


def test20_dict_iterator():
    assert t.test_10({}) == {}
    assert t.test_10({1: 2}) == {1: 2}
    assert t.test_10({1: 2, 3: 4}) == {1: 2, 3: 4}
    assert t.test_10({1: 2, 3: 4, "a": "b"}) == {1: 2, 3: 4, "a": "b"}


def test21_numpy_overloads():
    try:
        import numpy as np
    except ImportError:
        pytest.skip("numpy is missing")

    assert t.test_05(np.int32(0)) == 1
    assert t.test_05(np.float64(0.1)) == 2
    assert t.test_05(np.float64(0.0)) == 2
    assert t.test_11_sl(np.int32(5)) == 5
    assert t.test_11_ul(np.int32(5)) == 5
    assert t.test_11_sll(np.int32(5)) == 5
    assert t.test_11_ull(np.int32(5)) == 5

    with pytest.raises(TypeError) as excinfo:
        t.test_21_dnc(np.float64(21.0))  # Python type is not exactly float
    assert "incompatible function arguments" in str(excinfo.value)
    assert t.test_21_dnc(float(np.float64(21.0))) == 22.0
    assert t.test_21_dnc(float(np.float32(21.0))) == 22.0

    assert t.test_21_fnc(float(np.float32(21.0))) == 22.0
    with pytest.raises(TypeError) as excinfo:
        t.test_21_fnc(float(np.float64(21.1)))  # Inexact narrowing to float32
    assert "incompatible function arguments" in str(excinfo.value)
    assert t.test_21_fnc(float(np.float32(21.1))) == np.float32(22.1)


def test22_string_return():
    assert t.test_12("hello") == "hello"
    assert t.test_13() == "test"
    assert t.test_14("abc") == "abc"


def test23_byte_return():
    assert t.test_15(b"abc") == "abc"
    assert t.test_15_d(b"abc\x00def\x00ghi") == b"abc\x00def\x00ghi"
    assert t.test_16("hello") == b"hello"
    assert t.test_17(b"four") == 4
    assert t.test_17(b"\x00\x00\x00\x00") == 4
    assert t.test_18("hello world", 5) == b"hello"


@pytest.mark.skipif(sys.version_info < (3, 9), reason="requires python3.9 or higher")
def test24_pydoc():
    import pydoc

    assert "test_05(arg: int, /)" in pydoc.render_doc(t)


def test25_int():
    assert t.test_19(5) == 128
    assert t.test_20("5") == 128
    assert t.test_21(5) == 5
    assert t.test_21_f(5.1) == int(5.1)
    assert t.test_21_f(1e50) == int(1e50)
    assert type(t.test_21_f(0.5)) is int
    assert t.test_21_g() == int(1.5)
    assert type(t.test_21_g()) is int
    assert t.test_21_h() == int(1e50)
    assert type(t.test_21_h()) is int
    assert t.test_19.__doc__ == "test_19(arg: int, /) -> object"


def test26_capsule():
    p = t.test_22()
    assert "capsule" in str(p) and "nb_handle" in str(p)
    assert t.test_24(p) == 1
    p = t.test_23()
    assert p is None
    assert t.test_24(p) == 0


def test27_slice():
    s = slice(1, 10, 2)
    assert t.test_25(s) is s
    assert t.test_25.__doc__ == "test_25(arg: slice, /) -> slice"
    assert t.test_26() == slice(4)
    assert t.test_27() == slice(2, 10)
    assert t.test_28() == slice(5, -5, -2)


def test28_ellipsis():
    assert t.test_29(...) is ...
    assert (
        t.test_29.__doc__ == "test_29(arg: types.EllipsisType, /) -> types.EllipsisType"
    )


def test29_traceback():
    result = t.test_30(fail_fn)
    regexp = r'Traceback \(most recent call last\):\n.*\n  File "[^"]*", line [0-9]*, in fail_fn\n.*RuntimeError: Foo'
    matches = re.findall(regexp, result, re.MULTILINE | re.DOTALL)
    assert len(matches) == 1


def test30_noexcept():
    assert t.test_31(123) == 123
    assert t.test_32(123) == 123


@pytest.mark.parametrize(
    "func_name",
    [
        "identity_i8",
        "identity_u8",
        "identity_i16",
        "identity_u16",
        "identity_i32",
        "identity_u32",
        "identity_i64",
        "identity_u64",
    ],
)
def test31_range(func_name):
    func = getattr(t, func_name)

    values = [
        0,
        -1,
        1,
        2**7,
        2**7 - 1,
        2**8,
        2**8 - 1,
        2**15,
        2**15 - 1,
        2**16,
        2**16 - 1,
        2**29,
        2**29 - 1,
        2**30,
        2**30 - 1,
        2**31,
        2**31 - 1,
        2**32,
        2**32 - 1,
        2**63,
        2**63 - 1,
        2**64,
        2**64 - 1,
        2**127,
        2**127 - 1,
        2**128,
        2**128 - 1,
    ]
    values += [-value for value in values]
    suffix = func.__name__[9:]

    if suffix[0] == "u":
        range_min = 0
        range_max = 2 ** int(suffix[1:]) - 1
    else:
        range_min = -(2 ** (int(suffix[1:]) - 1))
        range_max = -range_min - 1

    for value in values:
        if value < range_min or value > range_max:
            with pytest.raises(TypeError):
                value_out = func(value)
        else:
            value_out = func(value)
            assert value_out == value


def test33_method_on_non_nanobind_class():
    class AClass:
        def __init__(self):
            self.x = 42

    AClass.simple_method = t.test_33
    AClass.complex_method = t.test_34
    a = AClass()
    assert a.simple_method(7) == 49
    assert a.complex_method(y=2) == 84


def test34_module_docstring():
    assert t.__doc__ == "function testcase"


def test35_return_capture():
    x = t.test_35()
    assert x() == "Test Foo"


def test36_test_char():
    assert t.test_cast_char("c") == "c"
    assert t.test_cast_char("\x00") == "\x00"
    with pytest.raises(TypeError):
        assert t.test_cast_char("abc")
    with pytest.raises(TypeError):
        assert t.test_cast_char("")
    with pytest.raises(RuntimeError):
        assert t.test_cast_char(123)


def test37_test_str():
    assert t.test_cast_str("c") == "c"
    assert t.test_cast_str("abc") == "abc"
    with pytest.raises(RuntimeError):
        assert t.test_cast_str(123)


def test38_set():
    x = t.test_set()
    assert isinstance(x, set)
    assert len(x) == 2
    assert 123 in x and "123" in x
    assert t.test_set_contains(x, 123)
    assert t.test_set_contains(x, "123")
    assert not t.test_set_contains(x, "1234")
    assert not t.test_set_contains(x, 1234)


def test39_del():
    l = [0, 1, 2, 3, 4]
    t.test_del_list(l)
    assert l == [0, 1, 3, 4]

    l = {"a": 0, "b": 1}
    t.test_del_dict(l)
    assert l == {"b": 1}

    with pytest.raises(IndexError):
        t.test_del_list([])

    with pytest.raises(KeyError):
        t.test_del_dict({})


def test40_nb_signature():
    assert t.test_01.__nb_signature__ == ((r"def test_01() -> None", None, None),)
    assert t.test_02.__nb_signature__ == (
        (r"def test_02(up: int = \0, down: int = \1) -> int", None, (8, 1)),
    )
    assert t.test_05.__nb_signature__ == (
        (r"def test_05(arg: int, /) -> int", "doc_1", None),
        (r"def test_05(arg: float, /) -> int", "doc_2", None),
    )
    if sys.version_info >= (3, 9):
        assert t.test_07.__nb_signature__ == (
            (
                r"def test_07(arg0: int, arg1: int, /, *args, **kwargs) -> tuple[int, int]",
                None,
                None,
            ),
            (
                r"def test_07(a: int, b: int, *myargs, **mykwargs) -> tuple[int, int]",
                None,
                None,
            ),
        )


def test41_kw_only():
    # (i, j, *args, z)
    assert t.test_args_kwonly(2, 2.5, z=22) == (2, 2.5, (), 22)
    assert t.test_args_kwonly(2, 2.5, "a", "b", z=22) == (2, 2.5, ("a", "b"), 22)
    assert t.test_args_kwonly(z=22, i=4, j=16) == (4, 16.0, (), 22)
    assert (
        t.test_args_kwonly.__doc__
        == "test_args_kwonly(i: int, j: float, *args, z: int) -> tuple"
    )
    with pytest.raises(TypeError):
        t.test_args_kwonly(2, 2.5, 22)  # missing z= keyword

    # (i, j, *args, z, **kwargs)
    assert t.test_args_kwonly_kwargs(i=1, k=4, j=10, z=-1, y=9) == (
        1, 10, (), -1, {"k": 4, "y": 9}
    )
    assert t.test_args_kwonly_kwargs(1, 2, 3, 4, z=11, y=12) == (
        1, 2, (3, 4), 11, {"y": 12}
    )
    with pytest.raises(TypeError):
        t.test_args_kwonly_kwargs(1, 2, 3, 4, 5)
    assert (
        t.test_args_kwonly_kwargs.__doc__
        == "test_args_kwonly_kwargs(i: int, j: float, *args, z: int, **kwargs) -> tuple"
    )

    # (i, *, j, **kwargs)
    assert t.test_kwonly_kwargs(j=2, i=1) == (1, 2, {})
    assert t.test_kwonly_kwargs(j=2, i=1, z=10) == (1, 2, {"z": 10})
    assert t.test_kwonly_kwargs(1, j=2) == (1, 2, {})
    assert t.test_kwonly_kwargs(1, j=2, z=10) == (1, 2, {"z": 10})
    with pytest.raises(TypeError):
        t.test_kwonly_kwargs(1, 2)
    with pytest.raises(TypeError):
        t.test_kwonly_kwargs(1, 2, j=3)
    with pytest.raises(TypeError):
        t.test_kwonly_kwargs(1, 2, z=10)
    assert (
        t.test_kwonly_kwargs.__doc__
        == "test_kwonly_kwargs(i: int, *, j: float, **kwargs) -> tuple"
    )

    # (*, i, j)
    assert t.test_kw_only_all(i=1, j=2) == (1, 2)
    assert t.test_kw_only_all(j=1, i=2) == (2, 1)
    with pytest.raises(TypeError):
        t.test_kw_only_all(i=1)
    with pytest.raises(TypeError):
        t.test_kw_only_all(1, 2)
    assert (
        t.test_kw_only_all.__doc__
        == "test_kw_only_all(*, i: int, j: int) -> tuple"
    )

    # (__arg0, *, j, k)
    assert t.test_kw_only_some(1, k=3, j=2) == (1, 2, 3)
    assert (
        t.test_kw_only_some.__doc__
        == "test_kw_only_some(arg0: int, *, j: int, k: int) -> tuple"
    )

    # (__arg0=3, j=4, *, k=5, z)
    assert t.test_kw_only_with_defaults(z=8) == (3, 4, 5, 8)
    assert t.test_kw_only_with_defaults(2, z=8) == (2, 4, 5, 8)
    assert t.test_kw_only_with_defaults(2, j=7, k=8, z=9) == (2, 7, 8, 9)
    assert t.test_kw_only_with_defaults(2, 7, z=9, k=8) == (2, 7, 8, 9)
    with pytest.raises(TypeError):
        t.test_kw_only_with_defaults(2, 7, 8, z=9)
    assert (
        t.test_kw_only_with_defaults.__doc__
        == "test_kw_only_with_defaults(arg0: int = 3, j: int = 4, *, k: int = 5, z: int) -> tuple"
    )

    # (i, *, j)
    assert t.test_kw_only_mixed(1, j=2) == (1, 2)
    assert t.test_kw_only_mixed(j=2, i=3) == (3, 2)
    assert t.test_kw_only_mixed(i=2, j=3) == (2, 3)
    with pytest.raises(TypeError):
        t.test_kw_only_mixed(i=1)
    with pytest.raises(TypeError):
        t.test_kw_only_mixed(1, i=2)
    assert (
        t.test_kw_only_mixed.__doc__
        == "test_kw_only_mixed(i: int, *, j: int) -> tuple"
    )

    with pytest.raises(TypeError):
        t.kw_only_methods(42)

    val = t.kw_only_methods(v=42)
    assert val.v == 42

    # (self, *, i, j)
    assert val.method_2k() == (1, 2)
    assert val.method_2k(i=3) == (3, 2)
    assert val.method_2k(j=4) == (1, 4)
    assert val.method_2k(i=3, j=4) == (3, 4)
    assert val.method_2k(j=3, i=4) == (4, 3)
    with pytest.raises(TypeError):
        val.method_2k(1)
    with pytest.raises(TypeError):
        val.method_2k(1, j=2)
    assert (
        t.kw_only_methods.method_2k.__doc__
        == "method_2k(self, *, i: int = 1, j: int = 2) -> tuple"
    )

    # (self, i, *, j)
    assert val.method_1p1k() == (1, 2)
    assert val.method_1p1k(i=3) == (3, 2)
    assert val.method_1p1k(j=4) == (1, 4)
    assert val.method_1p1k(i=3, j=4) == (3, 4)
    assert val.method_1p1k(j=3, i=4) == (4, 3)
    assert val.method_1p1k(3) == (3, 2)
    assert val.method_1p1k(3, j=4) == (3, 4)
    with pytest.raises(TypeError):
        val.method_2k(1, 2)
    assert (
        t.kw_only_methods.method_1p1k.__doc__
        == "method_1p1k(self, i: int = 1, *, j: int = 2) -> tuple"
    )


def test42_ptr_return():
    assert t.test_ptr_return() == (10, 100)

def test41_any():
    s = "hello"
    assert t.test_any(s) is s
    assert t.test_any.__doc__ == "test_any(arg: typing.Any, /) -> typing.Any"

def test42_wrappers_list():
    assert t.test_wrappers_list()

def test43_wrappers_dict():
    assert t.test_wrappers_dict()

def test43_wrappers_set():
    assert t.test_wrappers_set()

def test44_hash():
    value = (1, 2, 3)
    assert t.hash_it(value) == hash(value);


def test45_new():
    assert t.test_bytearray_new() == bytearray()
    assert t.test_bytearray_new("\x00\x01\x02\x03", 4) == bytearray(
        b"\x00\x01\x02\x03"
    )
    assert t.test_bytearray_new("", 0) == bytearray()


def test46_copy():
    o = bytearray(b"\x00\x01\x02\x03")
    c = t.test_bytearray_copy(o)
    assert c == o
    o.clear()
    assert c != o


def test47_c_str():
    o = bytearray(b"Hello, world!")
    assert t.test_bytearray_c_str(o) == "Hello, world!"


def test48_size():
    o = bytearray(b"Hello, world!")
    assert t.test_bytearray_size(o) == len(o)


def test49_resize():
    o = bytearray(b"\x00\x01\x02\x03")
    assert len(o) == 4
    t.test_bytearray_resize(o, 8)
    assert len(o) == 8


def test50_call_policy():
    def case(arg1, arg2, expect_ret):  # type: (str, str, str | None) -> str
        if hasattr(sys, "getrefcount"):
            refs_before = (sys.getrefcount(arg1), sys.getrefcount(arg2))

        ret = None
        try:
            ret = t.test_call_policy(arg1, arg2)
            assert ret == expect_ret
            return ret
        finally:
            if expect_ret is None:
                assert t.call_policy_record() == []
            else:
                (((arg1r, arg2r), recorded_ret),) = t.call_policy_record()
                assert recorded_ret == expect_ret
                assert ret is None or ret is recorded_ret
                assert recorded_ret is not expect_ret

                if hasattr(sys, "getrefcount"):
                    # Make sure no reference leak occurred: should be
                    # one in getrefcount args, one or two in locals,
                    # zero or one in the pending-return-value slot.
                    # We have to decompose this to avoid getting confused
                    # by transient additional references added by pytest's
                    # assertion rewriting.
                    ret_refs = sys.getrefcount(recorded_ret)
                    expected_refs = 2 + 2 * (ret is not None)

                    # On Python 3.14a7, an optimization was introduced where
                    # stack-based function calling no longer acquires a reference
                    if py_3_14a7_or_newer:
                        assert ret_refs == expected_refs - 1 or ret_refs == expected_refs
                    else:
                        assert ret_refs == expected_refs

                for (passed, recorded) in ((arg1, arg1r), (arg2, arg2r)):
                    if passed == "swapfrom":
                        assert recorded == "swapto"
                        if hasattr(sys, "getrefcount"):
                            recorded_refs = sys.getrefcount(recorded)

                            # recorded, arg1r, unnamed tuple, getrefcount arg
                            if py_3_14a7_or_newer:
                                assert recorded_refs == 3 or recorded_refs == 4
                            else:
                                assert recorded_refs == 4
                    else:
                        assert passed is recorded

                del passed, recorded, arg1r, arg2r
                if hasattr(sys, "getrefcount"):
                    refs_after = (sys.getrefcount(arg1), sys.getrefcount(arg2))

                    if not py_3_14a7_or_newer or ret is not None:
                        assert refs_before == refs_after

    # precall throws exception
    with pytest.raises(RuntimeError, match="expected only strings"):
        case(12345, "0", None)

    # conversion of args fails
    with pytest.raises(TypeError):
        case("string", "xxx", "<unfinished>")

    # function throws exception
    with pytest.raises(RuntimeError, match="offset too large"):
        case("abc", "4", "<unfinished>")

    # conversion of return value fails
    with pytest.raises(UnicodeDecodeError):
        case("returnfail", "4", "<return conversion failed>")

    # postcall throws exception
    with pytest.raises(RuntimeError, match="postcall exception"):
        case("postthrow", "4", "throw")

    # normal call
    case("example", "1", "xample")

    # precall modifies args
    case("swapfrom", "0", "swapto")
    with pytest.raises(TypeError):
        case("swapfrom", "xxx", "<unfinished>")
    with pytest.raises(RuntimeError, match="offset too large"):
        case("swapfrom", "10", "<unfinished>")

def test51_isinstance():
    assert t.isinstance_(3, int)
    assert not t.isinstance_(3, bool)
    with pytest.raises(TypeError):
        t.isinstance_(3, 7)


def test52_frozenset():
    x = t.test_frozenset()
    assert isinstance(x, frozenset)
    assert len(x) == 2
    assert 123 in x and "123" in x
    assert t.test_frozenset_contains(x, 123)
    assert t.test_frozenset_contains(x, "123")
    assert not t.test_frozenset_contains(x, "1234")
    assert not t.test_frozenset_contains(x, 1234)

def test53_fallback():
    assert t.test_fallback_1(3.0) == 0
    assert t.test_fallback_1(3) == 1
    assert t.test_fallback_1('3') == 1

    assert t.test_fallback_2(3.0) == 0
    assert t.test_fallback_2(3) == 0
    assert t.test_fallback_2('3') == 1

def test54_dict_default():
    assert t.test_get_dict_default({'key': 100}) == 100
    assert t.test_get_dict_default({'key2': 100}) == 123



#!/usr/bin/env python3
"""
Test for WRP CTE Core Python query API bindings
Tests TagQuery and BlobQuery methods
"""

import sys
import os

# When running with python -I (isolated mode), we need to manually add the current directory
# The test is run with WORKING_DIRECTORY set to the module directory
sys.path.insert(0, os.getcwd())

def test_import():
    """Test that the module can be imported successfully"""
    try:
        import wrp_cte_core_ext as cte
        print("✅ Module import successful")
        return True
    except ImportError as e:
        print(f"❌ Module import failed: {e}")
        return False

def test_query_methods_exist():
    """Test that TagQuery and BlobQuery methods exist on Client class"""
    try:
        import wrp_cte_core_ext as cte

        # Test Client type accessibility
        client_type = cte.Client
        print(f"✅ Client type accessible: {client_type}")

        # Check if TagQuery method exists
        if hasattr(client_type, 'TagQuery'):
            print("✅ TagQuery method exists on Client class")
        else:
            print("❌ TagQuery method not found on Client class")
            return False

        # Check if BlobQuery method exists
        if hasattr(client_type, 'BlobQuery'):
            print("✅ BlobQuery method exists on Client class")
        else:
            print("❌ BlobQuery method not found on Client class")
            return False

        return True
    except Exception as e:
        print(f"❌ Query methods test failed: {e}")
        return False

def test_query_method_signatures():
    """Test that query methods have correct signatures"""
    try:
        import wrp_cte_core_ext as cte
        import inspect

        client_type = cte.Client

        # Check TagQuery signature
        if hasattr(client_type, 'TagQuery'):
            tag_query_method = getattr(client_type, 'TagQuery')
            print(f"✅ TagQuery method signature accessible")
            # Note: nanobind methods don't always expose full signature info
            print(f"   TagQuery: {tag_query_method}")

        # Check BlobQuery signature
        if hasattr(client_type, 'BlobQuery'):
            blob_query_method = getattr(client_type, 'BlobQuery')
            print(f"✅ BlobQuery method signature accessible")
            print(f"   BlobQuery: {blob_query_method}")

        return True
    except Exception as e:
        print(f"❌ Method signature test failed: {e}")
        return False

def test_query_with_mock_client():
    """Test query methods accessibility without full runtime"""
    try:
        import wrp_cte_core_ext as cte

        # Create MemContext (doesn't require runtime)
        mctx = cte.MemContext()
        print("✅ Created MemContext for query calls")

        # Test that we can access client methods (won't call without runtime)
        client_type = cte.Client

        # Verify methods are callable (signature check)
        tag_query = getattr(client_type, 'TagQuery', None)
        blob_query = getattr(client_type, 'BlobQuery', None)

        if tag_query and callable(tag_query):
            print("✅ TagQuery method is callable")
        else:
            print("❌ TagQuery method is not callable")
            return False

        if blob_query and callable(blob_query):
            print("✅ BlobQuery method is callable")
        else:
            print("❌ BlobQuery method is not callable")
            return False

        print("✅ Query methods binding verification successful")
        return True
    except Exception as e:
        print(f"❌ Query method accessibility test failed: {e}")
        return False

def test_query_parameter_types():
    """Test that query methods accept expected parameter types"""
    try:
        import wrp_cte_core_ext as cte

        # Create required types
        mctx = cte.MemContext()
        print("✅ Created MemContext")

        # Test string parameter (for regex patterns)
        tag_regex = "user_.*"
        blob_regex = "blob_.*"
        print(f"✅ Created regex patterns: '{tag_regex}', '{blob_regex}'")

        # Note: We can't actually call the methods without full runtime initialization,
        # but we've verified the types can be created and the methods exist
        print("✅ Parameter types verified for query methods")

        return True
    except Exception as e:
        print(f"❌ Parameter types test failed: {e}")
        return False

def test_query_documentation():
    """Test that query methods have documentation"""
    try:
        import wrp_cte_core_ext as cte

        client_type = cte.Client

        # Check TagQuery documentation
        tag_query = getattr(client_type, 'TagQuery', None)
        if tag_query and hasattr(tag_query, '__doc__') and tag_query.__doc__:
            print(f"✅ TagQuery has documentation: {tag_query.__doc__}")
        else:
            print("⚠️  TagQuery documentation not found (may be expected)")

        # Check BlobQuery documentation
        blob_query = getattr(client_type, 'BlobQuery', None)
        if blob_query and hasattr(blob_query, '__doc__') and blob_query.__doc__:
            print(f"✅ BlobQuery has documentation: {blob_query.__doc__}")
        else:
            print("⚠️  BlobQuery documentation not found (may be expected)")

        return True
    except Exception as e:
        print(f"❌ Documentation test failed: {e}")
        return False

def main():
    """Run all query API binding tests"""
    print("🧪 Running Python query API bindings tests...")

    tests = [
        ("Module Import", test_import),
        ("Query Methods Exist", test_query_methods_exist),
        ("Query Method Signatures", test_query_method_signatures),
        ("Query Methods Accessibility", test_query_with_mock_client),
        ("Query Parameter Types", test_query_parameter_types),
        ("Query Documentation", test_query_documentation)
    ]

    passed = 0
    total = len(tests)

    for test_name, test_func in tests:
        print(f"\n📋 Testing {test_name}...")
        if test_func():
            passed += 1
        else:
            print(f"❌ {test_name} failed")

    print(f"\n📊 Test Results: {passed}/{total} tests passed")

    if passed == total:
        print("🎉 All query API binding tests passed!")
        return 0
    else:
        print("💥 Some tests failed!")
        return 1

if __name__ == "__main__":
    sys.exit(main())



"""
Prepends the preabmle to all files in the repo

USAGE:
    python3 scripts/preamble.py ${HSHM_ROOT}
"""

import sys,os,re
import pathlib

hermes_preamble = """
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Distributed under BSD 3-Clause license.                                   *
 * Copyright by The HDF Group.                                               *
 * Copyright by the Illinois Institute of Technology.                        *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of Hermes. The full Hermes copyright notice, including  *
 * terms governing use, modification, and redistribution, is contained in    *
 * the COPYING file, which can be found at the top directory. If you do not  *
 * have access to the file, you may request a copy from help@hdfgroup.org.   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
""".strip()

preamble2 = """
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
* Distributed under BSD 3-Clause license.                                   *
* Copyright by The HDF Group.                                               *
* Copyright by the Illinois Institute of Technology.                        *
* All rights reserved.                                                      *
*                                                                           *
* This file is part of Hermes. The full Hermes copyright notice, including  *
* terms governing use, modification, and redistribution, is contained in    *
* the COPYING file, which can be found at the top directory. If you do not  *
* have access to the file, you may request a copy from help@hdfgroup.org.   *
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
""".strip()

labstor_preamble = """
/*
 * Copyright (C) 2022  SCS Lab <scslab@iit.edu>,
 * Luke Logan <llogan@hawk.iit.edu>,
 * Jaime Cernuda Garcia <jcernudagarcia@hawk.iit.edu>
 * Jay Lofstead <gflofst@sandia.gov>,
 * Anthony Kougkas <akougkas@iit.edu>,
 * Xian-He Sun <sun@iit.edu>
 *
 * This file is part of HermesShm
 *
 * HermesShm is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/>.
 */
""".strip()

def PrependPreamble(path):
    text = ""
    with open(path, 'r') as fp:
        text = fp.read()
        text = text.replace(hermes_preamble, "")
        text = text.replace(preamble2, "")
        text = text.replace(labstor_preamble, "")
        text = re.sub("//\n// Created by [^\n]*\n//\n", "", text)
        text = text.strip()
        text = hermes_preamble + "\n\n" + text + "\n"

    with open(path, 'w') as fp:
        fp.write(text)


def LocateCppFiles(root):
    for entry in os.listdir(root):
        full_path = os.path.join(root, entry)
        if os.path.isdir(full_path):
            LocateCppFiles(full_path)
        elif os.path.isfile(full_path):
            file_ext = pathlib.Path(full_path).suffix
            if file_ext == '.cc' or file_ext == '.h':
                PrependPreamble(full_path)


PROJECT_ROOT=pathlib.Path(__file__).parent.parent.resolve()
LocateCppFiles(PROJECT_ROOT)


"""
IOWarp Core - High-performance distributed I/O and task execution runtime.

This package provides Python bindings and installation support for the IOWarp
ecosystem, which includes:

- Context Transport Primitives (HermesShm): Shared memory IPC primitives
- Runtime (Chimaera): Distributed task execution runtime
- Context Transfer Engine (Hermes): Multi-tiered I/O buffering system
- Context Assimilation Engine: Data ingestion and processing engine

All components are built from C++ source using CMake during installation.
"""

try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:
    from importlib_metadata import version, PackageNotFoundError

try:
    __version__ = version("iowarp-core")
except PackageNotFoundError:
    # Package is not installed, use a fallback version
    __version__ = "0.0.0.dev0"

__author__ = "IOWarp Team"
__license__ = "BSD-3-Clause"

# Component versions and information
COMPONENTS = {
    "context-transport-primitives": {
        "description": "High-performance shared memory library with IPC primitives",
        "url": "https://github.com/iowarp/context-transport-primitives",
    },
    "runtime": {
        "description": "Distributed task execution runtime (Chimaera)",
        "url": "https://github.com/iowarp/runtime",
    },
    "context-transfer-engine": {
        "description": "Multi-tiered I/O buffering system (Hermes)",
        "url": "https://github.com/iowarp/context-transfer-engine",
    },
    "context-assimilation-engine": {
        "description": "Data ingestion and processing engine",
        "url": "https://github.com/iowarp/context-assimilation-engine",
    },
}


def get_component_info():
    """Get information about installed IOWarp components."""
    return COMPONENTS


def get_version():
    """Get the version of iowarp-core package."""
    return __version__


__all__ = [
    "__version__",
    "COMPONENTS",
    "get_component_info",
    "get_version",
]



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

"""
Matplotlib window with toolbar.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
import wx
try:  # for GeoArray we use cartopy that can be challenging to freeze on OSX to dependencies (i.e. geos)
    import cartopy.crs as ccrs
    from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER
except (ImportError, OSError):
    pass
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg as FigCanvas
from matplotlib.backends.backend_wx import NavigationToolbar2Wx as NavigationToolbar

import logging
log = logging.getLogger(__name__)

from ..frame import BaseFrame

ID_VIEW_CMAP_JET = wx.ID_ANY  # default
ID_VIEW_CMAP_BONE = wx.ID_ANY
ID_VIEW_CMAP_GIST_EARTH = wx.ID_ANY
ID_VIEW_CMAP_OCEAN = wx.ID_ANY
ID_VIEW_CMAP_RAINBOW = wx.ID_ANY
ID_VIEW_CMAP_RDYLGN = wx.ID_ANY
ID_VIEW_CMAP_WINTER = wx.ID_ANY
ID_SAVE_AS = wx.ID_ANY
ID_COPY = wx.ID_ANY
ID_ZOOM_IN = wx.ID_ANY
ID_ZOOM_OUT = wx.ID_ANY
ID_ZOOM_FIT = wx.ID_ANY


class PlotFrame(BaseFrame):
    """ Base class for Matplotlib plot windows.

    Override draw_figure() to plot your figure on the provided axes.
    """

    def __init__(self, data, title="a title"):
        """ Create a new Matplotlib plotting window for a 1D line plot """

        log.debug(self.__class__.__name__)
        super(PlotFrame, self).__init__(id=wx.ID_ANY, title=title, size=(800, 400))

        self.data = data

        self.panel = wx.Panel(self)

        self.dpi = 100
        self.fig = Figure((6.0, 4.0), dpi=self.dpi)
        self.canvas = FigCanvas(self.panel, -1, self.fig)

        self.axes = self.fig.add_subplot(111, projection=ccrs.PlateCarree())
        self.toolbar = NavigationToolbar(self.canvas)

        self.vbox = wx.BoxSizer(wx.VERTICAL)
        self.vbox.Add(self.canvas, 1, wx.LEFT | wx.TOP | wx.GROW)
        self.vbox.Add(self.toolbar, 0, wx.EXPAND)

        self.panel.SetSizer(self.vbox)
        self.vbox.Fit(self)

        # Add standard toolbar buttons
        self.toolbar.AddTool(wx.ID_SAVE, "Save", 
                            wx.ArtProvider.GetBitmap(wx.ART_FILE_SAVE, wx.ART_TOOLBAR),
                            "Save plot to file")
        
        self.toolbar.AddTool(wx.ID_COPY, "Copy", 
                            wx.ArtProvider.GetBitmap(wx.ART_COPY, wx.ART_TOOLBAR),
                            "Copy plot to clipboard")
        
        self.toolbar.AddTool(wx.ID_ZOOM_IN, "Zoom In", 
                            wx.ArtProvider.GetBitmap(wx.ART_ADD_BOOKMARK, wx.ART_TOOLBAR),
                            "Zoom in")
        
        self.toolbar.AddTool(wx.ID_ZOOM_OUT, "Zoom Out", 
                            wx.ArtProvider.GetBitmap(wx.ART_DEL_BOOKMARK, wx.ART_TOOLBAR),
                            "Zoom out")
        
        self.toolbar.AddTool(wx.ID_ZOOM_FIT, "Fit", 
                            wx.ArtProvider.GetBitmap(wx.ART_FIND, wx.ART_TOOLBAR),
                            "Fit plot to window")
        
        self.toolbar.Realize()

        # Bind events
        self.Bind(wx.EVT_TOOL, self.on_save, id=wx.ID_SAVE)
        self.Bind(wx.EVT_TOOL, self.on_copy, id=wx.ID_COPY)
        self.Bind(wx.EVT_TOOL, self.on_zoom_in, id=wx.ID_ZOOM_IN)
        self.Bind(wx.EVT_TOOL, self.on_zoom_out, id=wx.ID_ZOOM_OUT)
        self.Bind(wx.EVT_TOOL, self.on_zoom_fit, id=wx.ID_ZOOM_FIT)

        self.draw_figure()

    def draw_figure(self):
        raise NotImplementedError


class LinePlotFrame(PlotFrame):
    def __init__(self, data, names=None, title="Line Plot"):
        self.names = names
        PlotFrame.__init__(self, data, title)

    def draw_figure(self):

        lines = [self.axes.plot(d)[0] for d in self.data]
        if self.names is not None:
            for n in self.names:
                self.axes.legend(tuple(lines), tuple(self.names))


class ContourPlotFrame(PlotFrame):
    def __init__(self, data, extent, names=None, title="Contour Map"):
        self.geo_extent = extent
        log.debug("Extent: %f, %f, %f, %f" % self.geo_extent)
        # need to be set before calling the parent (need for plotting)
        self.colormap = "jet"
        self.cb = None  # matplotlib color-bar
        self.surf = None
        self.xx = None
        self.yy = None

        PlotFrame.__init__(self, data, title)

        self.cmap_menu = wx.Menu()
        self.cmap_menu.Append(ID_VIEW_CMAP_JET, "Jet", kind=wx.ITEM_RADIO)
        self.cmap_menu.Append(ID_VIEW_CMAP_BONE, "Bone", kind=wx.ITEM_RADIO)
        self.cmap_menu.Append(ID_VIEW_CMAP_GIST_EARTH, "Gist Earth", kind=wx.ITEM_RADIO)
        self.cmap_menu.Append(ID_VIEW_CMAP_OCEAN, "Ocean", kind=wx.ITEM_RADIO)
        self.cmap_menu.Append(ID_VIEW_CMAP_RAINBOW, "Rainbow", kind=wx.ITEM_RADIO)
        self.cmap_menu.Append(ID_VIEW_CMAP_RDYLGN, "Red-Yellow-Green", kind=wx.ITEM_RADIO)
        self.cmap_menu.Append(ID_VIEW_CMAP_WINTER, "Winter", kind=wx.ITEM_RADIO)
        self.add_menu(self.cmap_menu, "Colormap")

        self.Bind(wx.EVT_MENU, self.on_cmap_jet, id=ID_VIEW_CMAP_JET)
        self.Bind(wx.EVT_MENU, self.on_cmap_bone, id=ID_VIEW_CMAP_BONE)
        self.Bind(wx.EVT_MENU, self.on_cmap_gist_earth, id=ID_VIEW_CMAP_GIST_EARTH)
        self.Bind(wx.EVT_MENU, self.on_cmap_ocean, id=ID_VIEW_CMAP_OCEAN)
        self.Bind(wx.EVT_MENU, self.on_cmap_rainbow, id=ID_VIEW_CMAP_RAINBOW)
        self.Bind(wx.EVT_MENU, self.on_cmap_rdylgn, id=ID_VIEW_CMAP_RDYLGN)
        self.Bind(wx.EVT_MENU, self.on_cmap_winter, id=ID_VIEW_CMAP_WINTER)

        self.status_bar = wx.StatusBar(self, -1)
        self.status_bar.SetFieldsCount(2)
        self.SetStatusBar(self.status_bar)

        self.canvas.mpl_connect('motion_notify_event', self.update_status_bar)
        self.canvas.Bind(wx.EVT_ENTER_WINDOW, self.change_cursor)

    def on_cmap_jet(self, evt):
        log.debug("cmap: jet")
        self.colormap = "jet"
        self._refresh_plot()

    def on_cmap_bone(self, evt):
        log.debug("cmap: bone")
        self.colormap = "bone"
        self._refresh_plot()

    def on_cmap_gist_earth(self, evt):
        log.debug("cmap: gist_earth")
        self.colormap = "gist_earth"
        self._refresh_plot()

    def on_cmap_ocean(self, evt):
        log.debug("cmap: ocean")
        self.colormap = "ocean"
        self._refresh_plot()

    def on_cmap_rainbow(self, evt):
        log.debug("cmap: rainbow")
        self.colormap = "rainbow"
        self._refresh_plot()

    def on_cmap_rdylgn(self, evt):
        log.debug("cmap: RdYlGn")
        self.colormap = "RdYlGn"
        self._refresh_plot()

    def on_cmap_winter(self, evt):
        log.debug("cmap: winter")
        self.colormap = "winter"
        self._refresh_plot()

    def _refresh_plot(self):
        self.draw_figure()
        self.canvas.draw()

    def draw_figure(self):
        max_elements = 500  # don't attempt plot more than 500x500 elements
        rows = self.data.shape[0]
        cols = self.data.shape[1]
        row_stride = rows // max_elements + 1
        col_stride = cols // max_elements + 1
        self.surf = self.data[::row_stride, ::col_stride]
        self.xx = np.linspace(self.geo_extent[0], self.geo_extent[1], self.surf.shape[1])
        self.yy = np.linspace(self.geo_extent[2], self.geo_extent[3], self.surf.shape[0])
        img = self.axes.contourf(self.xx, self.yy, self.surf, 25, cmap=plt.cm.get_cmap(self.colormap),
                                 transform=ccrs.PlateCarree())
        self.axes.coastlines(resolution='50m', color='gray', linewidth=1)
        # add gridlines with labels only on the left and on the bottom
        grl = self.axes.gridlines(crs=ccrs.PlateCarree(), color='gray', draw_labels=True)
        grl.xformatter = LONGITUDE_FORMATTER
        grl.yformatter = LATITUDE_FORMATTER
        grl.xlabel_style = {'size': 8}
        grl.ylabel_style = {'size': 8}
        grl.ylabels_right = False
        grl.xlabels_top = False

        if self.cb:
            self.cb.on_mappable_changed(img)
        else:
            self.cb = plt.colorbar(img, ax=self.axes)
        self.cb.ax.tick_params(labelsize=8)

    def change_cursor(self, event):
        self.canvas.SetCursor(wx.StockCursor(wx.CURSOR_CROSS))

    @staticmethod
    def _find_nearest(arr, value):
        """ Helper function to find the nearest value in an array """
        return (np.abs(arr - value)).argmin()

    def update_status_bar(self, event):
        msg = str()
        if event.inaxes:
            x, y = event.xdata, event.ydata
            id_y, id_x = self._find_nearest(self.yy, y), self._find_nearest(self.xx, x)
            # log.debug("id: %f %f" % (id_y, id_x))
            z = self.surf[id_y, id_x]
            msg = "x= %f, y= %f, z= %f" % (x, y, z)
        self.status_bar.SetStatusText(msg, 1)



import test_callbacks_ext as t
import gc


def test_callbacks():
    pub1 = t.publisher()
    pub2 = t.publisher()
    record = []

    def sub1(x):
        record.append(x + 10)

    def sub2(x):
        record.append(x + 20)

    pub1.subscribe(sub1)
    pub2.subscribe(sub2)
    for pub in (pub1, pub2):
        pub.subscribe(record.append)

    pub1.emit(1)
    assert record == [11, 1]
    del record[:]

    pub2.emit(2)
    assert record == [22, 2]
    del record[:]

    pub1_w, pub2_w = t.registry.keys()  # weakrefs to pub1, pub2
    assert pub1_w() is pub1
    assert pub2_w() is pub2
    assert t.registry[pub1_w].subscribers == {sub1, record.append}
    assert t.registry[pub2_w].subscribers == {sub2, record.append}

    # NB: this `record.append` is a different object than the one we subscribed
    # above, so we're testing the normalization logic in unsubscribe_policy
    pub1.unsubscribe(record.append)
    assert t.registry[pub1_w].subscribers == {sub1}
    pub1.emit(3)
    assert record == [13]
    del record[:]

    del pub, pub1
    gc.collect()
    gc.collect()
    assert pub1_w() is None
    assert pub2_w() is pub2
    assert t.registry.keys() == {pub2_w}

    pub2.emit(4)
    assert record == [24, 4]
    del record[:]

    del pub2
    gc.collect()
    gc.collect()
    assert pub2_w() is None
    assert not t.registry



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

"""
Implementation of compass_model classes for HDF5 REST API.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from itertools import groupby
import sys
import os.path as op
import posixpath as pp
import json
import requests
import numpy as np

import logging
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

# Py2App can't successfully import otherwise
from hdf_compass import compass_model
from hdf_compass.utils import url2path

from . import hdf5dtype


def get_json(endpoint, domain=None, uri=None):
               
    # try to do a GET from the domain
    req = endpoint
    if uri is not None:
        req += uri

    headers = {}
    if domain is not None:        
        headers['host'] = domain
    
    log.debug("GET: " + req)
     
    rsp = requests.get(req, headers=headers, verify=False)
    log.debug("RSP: " + str(rsp.status_code) + ':' + rsp.text)
    
    if rsp.status_code != 200:
        raise IOError(rsp.reason)  
    rsp_json = json.loads(rsp.text)
                    
    return rsp_json


def sort_key(name):
    """ Sorting key for names in an HDF5 group.

    We provide "natural" sort order; e.g. "7" comes before "12".
    """
    return [(int(''.join(g)) if k else ''.join(g)) for k, g in groupby(name, key=unicode.isdigit)]


class HDF5RestStore(compass_model.Store):
    """
    Data store implementation for an HDF Service endpoint

    Keys are valid h5paths to the object:
        e.g.:
            /g1
            /g1/dset1.1
            /g1/dtype
            
    Values are URI's to the resource
            /groups/<uuid>
            /datasets/<uuid>
            /datatypes/<uuid>
    """
    @staticmethod
    def plugin_name():
        return "HDF5 Rest"

    @staticmethod
    def plugin_description():
        return "A plugin used to access HDF Services."

    def __contains__(self, key):
        if key in self.f:
            return True
        try:
            pkey = self.get_parent(key)
        except KeyError:
            return False
        if pkey not in self.f:
            # assumes that we've loaded any parents, before querying for children
            return False
        pkey_uri = self.f[pkey]
        if not pkey_uri.startswith("/groups/"):
            # if the parent is not a group, this is not a valid key
            return False
        # do a get on the link name
        linkname = pp.basename(key)
        contains = False
        try:
            link_json = self.get(pkey_uri + "/links/" + linkname)
            if link_json["class"] == "H5L_TYPE_HARD":
                log.debug("add key to store:" + key)
                self.f[key] = '/' + link_json["collection"] + '/' + link_json["id"]
                contains = True
            else:
                pass # todo support soft/external links
            
        except IOError:
            # invalid link
            # todo - verify it is a 404
            log.debug("invalid key:"+key)
        return contains
        
     

    @property
    def url(self):
        return self._url

    @property
    def display_name(self):
        if self.domain:
            return self.domain
        else:
            return self.endpoint
         
    @property
    def root(self):
        return self['/']

    @property
    def valid(self):
        return '/' in self.f
        

    @staticmethod
    def can_handle(url):
        log.debug("hdf5rest can_handle: " + url)
        try:
            flag = True
            rsp_json = get_json(url)
            for key in ("root", "created", "hrefs", "lastModified"):
                if key not in rsp_json:
                    flag = False
                    break
            log.debug("able to handle %s? %r" % (url, flag))
            return flag
        except Exception:
            log.debug("able to handle %s? no" % url)
            return False
        return True
        

    def __init__(self, url):
        if not self.can_handle(url):
            raise ValueError(url)
            
        self._url = url
        # extract domain if there's a "host" query param
        queryParam = "host="
       
        nindex = url.find('?' + queryParam)
        if nindex < 0:
            nindex = url.find('&' + queryParam)
        if nindex > 0:
            domain = url[(nindex + len(queryParam) + 1):]
            # trim any additional query params
            nindex = domain.find('&')
            if nindex > 0:
                domain = domain[:nindex]
            self._domain = domain
        else:
            self._domain = None
            
        nindex = url.find('?')
        if nindex < 0:
            self._endpoint = url
        else:
            self._endpoint = url[:(nindex)]
            if self._endpoint.endswith('/'):
                # trim any trailing '/'
                self._endpoint = self._endpoint[:-1]
         
        self.cache = {} # http cache    
        rsp = self.get('/')
        
        self.f = {}
        self.f['/'] = "/groups/" + rsp['root']
        
        
    @property
    def endpoint(self):
        return self._endpoint
        
    @property
    def domain(self):
        return self._domain
        
    @property
    def objid(self):
        return self._objid
        
    def get(self, uri):
        
        if uri in self.cache:
            rsp = self.cache[uri]
        else:
            rsp = get_json(self.endpoint, domain=self.domain, uri=uri)
            self.cache[uri] = rsp
        return rsp
        
    def close(self):
        self.f = {}  # clear the key store

    def get_parent(self, key):
        # HDFCompass requires the parent of the root container be None
        if key == "" or key == "/":
            return None
            
        pkey = pp.dirname(key)
        if pkey == "":
            pkey = "/"
            
        if pkey not in self.f:
            # is it possible to get to a key without traversing the parents?
            # if so, we should query the server for the given path
            raise KeyError("parent not found: " + pkey)
        return self[pkey]


class HDF5RestGroup(compass_model.Container):
    """ Represents an HDF5 group, to be displayed in the browser view. """

    class_kind = "HDF5 Group"

    @staticmethod
    def can_handle(store, key):
        return key in store and store.f[key].startswith("/groups/")

    def get_names(self):

        # Lazily build the list of names; this helps when browsing big files
        if self._xnames is None:
            rsp = self.store.get(self._uri + "/links")
            self._xnames = []
            links = rsp["links"]
            log.debug("got %d links for key: %s" % (len(links), self._key))
            for link in links:
                name = link["title"]
                self._xnames.append(name)
                link_key = pp.join(self.key, name)
                if link_key not in self.store.f:
                    if link["class"] == "H5L_TYPE_HARD":
                        log.debug("add key to store:" + link_key)
                        self.store.f[link_key] = '/' + link["collection"] + '/' + link["id"]
                    else:
                        pass # todo support soft/external links

            # Natural sort is expensive
            if len(self._xnames) < 1000:
                self._xnames.sort(key=sort_key)

        return self._xnames

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self._uri = store.f[key]
        self._xnames = None
        rsp = store.get(self._uri)
        self._count = rsp["linkCount"]
        log.debug("new group node: " + self._key)
        self.get_names()

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        name = pp.basename(self.key)
        if name == "":
            name = '/'
        return name

    @property
    def display_title(self):
        return "%s %s" % (self.store.display_name, self.key)

    @property
    def description(self):
        return 'Group "%s" (%d members)' % (self.display_name, len(self))

    def __len__(self):
        return self._count

    def __iter__(self):
        for name in self._xnames:
            yield self.store[pp.join(self.key, name)]

    def __getitem__(self, idx):
        name = self._xnames[idx]
        return self.store[pp.join(self.key, name)]


class HDF5RestDataset(compass_model.Array):
    """ Represents an HDF5 dataset. """

    class_kind = "HDF5 Dataset"

    @staticmethod
    def can_handle(store, key):
        return key in store and store.f[key].startswith("/datasets/")

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self._uri = store.f[key]
        rsp = store.get(self._uri)
        shape_json = rsp["shape"]
        if shape_json["class"]  == "H5S_SCALAR":
            self._shape = ()
        elif shape_json["class"] == "H5S_SIMPLE":
            self._shape = shape_json["dims"]
        else:
            raise IOError("Unexpected shape class: " + shape_json["class"])
        type_json = rsp["type"]
        self._dtype = hdf5dtype.createDataType(type_json)
            

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return pp.basename(self.key)

    @property
    def description(self):
        return 'Dataset "%s"' % (self.display_name,)

    @property
    def shape(self):
        return self._shape

    @property
    def dtype(self):
        return self._dtype

    def __getitem__(self, args):
         log.debug("getitem: " + str(args))
         req = self._uri + "/value"
         rank = len(self._shape)
         if rank > 0:
             sel_query = '['
             for dim in range(rank):
                 if dim < len(args):
                     s = args[dim]
                 else:
                     s = slice(0, self._shape[dim])
                 sel_query += str(s.start)
                 sel_query += ':'
                 if s.stop > self._shape[dim]:
                     sel_query += str(self._shape[dim])
                 else:
                    sel_query += str(s.stop)
                 sel_query += ','
             sel_query = sel_query[:-1]  # trim trailing comma
             sel_query += ']'
             req += "?select=" + sel_query
         
             
         rsp = self.store.get(req)
         value = rsp["value"]
         arr = np.array(value, dtype=self._dtype)
         return arr
               

    def is_plottable(self):
        if self.dtype.kind == 'S':
            log.debug("Not plottable since ASCII String (characters: %d)" % self.dtype.itemsize)
            return False
        if self.dtype.kind == 'U':
            log.debug("Not plottable since Unicode String (characters: %d)" % self.dtype.itemsize)
            return False
        return True


class HDF5RestKV(compass_model.KeyValue):
    """ A KeyValue node used for HDF5 attributes. """

    class_kind = "HDF5 Attributes"

    @staticmethod
    def can_handle(store, key):
        canhandle = False
        if key in store:
            uri = store.f[key]
            if uri.startswith("/groups/"):
                canhandle = True
            elif uri.startswith("/datasets/"):
                canhandle = True
            elif uri.startswith("/datatypes/"):
                canhandle = True
        return canhandle

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self._uri = store.f[key]
        
        rsp = store.get(self._uri + "/attributes")
        attributes = rsp["attributes"]
        names = []
        for attr in attributes:
            names.append(attr["name"])
        self._names = names

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        n = pp.basename(self.key)
        return n if n != '' else '/'

    @property
    def description(self):
        return self.display_name

    @property
    def keys(self):
        return self._names[:]

    def __getitem__(self, name):
        rsp = self._store.get(self._uri + "/attributes/" + name)
        type_json = rsp["type"]
        value_json = rsp["value"]
        arr_dtype = hdf5dtype.createDataType(type_json)
        arr = np.array(value_json, dtype=arr_dtype)
        return arr




# Register handlers    
HDF5RestStore.push(HDF5RestKV)
HDF5RestStore.push(HDF5RestDataset)
#HDF5RestStore.push(HDF5Text)
HDF5RestStore.push(HDF5RestGroup)
#HDF5RestStore.push(HDF5Image)

compass_model.push(HDF5RestStore)




"""
Create the .h and .cc files for defining hermes singletons
USAGE:
    cd /path/to/hermes
    python3 scripts/singleton_generator.py
"""

import re

def ToCamelCase(string):
    if string is None:
        return
    words = re.sub(r"(_|-)+", " ", string).split()
    words = [word.capitalize() for word in words]
    return "".join(words)

def ToSnakeCase(string):
    if string is None:
        return
    string = re.sub('(\.|-)+', '_', string)
    words = re.split('([A-Z][^A-Z]*)', string)
    words = [word for word in words if len(word)]
    string = "_".join(words)
    return string.lower()



import os
import pathlib
import difflib
import sys
import platform
import pytest

is_unsupported = platform.python_implementation() == 'PyPy' or sys.version_info < (3, 10)
skip_on_unsupported = pytest.mark.skipif(
    is_unsupported, reason="Stub generation is only tested on CPython >= 3.10.0")

def remove_platform_dependent(s):
    '''Remove platform-dependent functions from the stubs'''
    s2 = []
    i = 0
    while i < len(s):
        v = s[i]
        if v.strip().startswith('float16'):
            i += 1
        elif v.startswith('def ret_numpy_half()') or \
           v.startswith('def test_slots()') or \
           v.startswith('TypeAlias'):
            i += 2
        else:
            s2.append(v)
            i += 1
    return s2


ref_paths = list(pathlib.Path(__file__).parent.rglob('*.pyi.ref'))
ref_path_ids = [p.name[:-len('.pyi.ref')] for p in ref_paths]
assert len(ref_paths) > 0, "Stub reference files not found!"

@skip_on_unsupported
@pytest.mark.parametrize('p_ref', ref_paths, ids=ref_path_ids)
def test01_check_stub_refs(p_ref, request):
    """
    Check that generated stub files match reference input
    """
    if not request.config.getoption('enable-slow-tests') and any(
        (x in p_ref.name for x in ['jax', 'tensorflow'])):
        pytest.skip("skipping because slow tests are not enabled")

    p_in = p_ref.with_suffix('')
    with open(p_ref, 'r') as f:
        s_ref = f.read().split('\n')
    with open(p_in, 'r') as f:
        s_in = f.read().split('\n')

    if "test_functions_ext" in p_in.name and sys.version_info < (3, 13):
        s_ref = [line.replace("types.CapsuleType", "typing_extensions.CapsuleType") for line in s_ref]
        s_ref.insert(3, "")
        s_ref.insert(4, "import typing_extensions")

    s_in = remove_platform_dependent(s_in)
    s_ref = remove_platform_dependent(s_ref)

    diff = list(difflib.unified_diff(
        s_ref,
        s_in,
        fromfile=str(p_ref),
        tofile=str(p_in)
    ))
    if len(diff):
        for p in diff:
            print(p.rstrip(), file=sys.stderr)
        print(
            '\nWarning: generated stubs do not match their references. If you\n'
            'intentionally changed a test suite extension, it may be necessary\n'
            'to replace the .pyi.ref file with the generated .pyi file. But\n'
            'please double-check that the change makes sense.',
            file=sys.stderr
        )
        assert False



import os
import sys
import logging

# Set up logging
logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)

# Add the project root to the Python path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__))))

from hdf_compass.compass_model.test import container, store
from hdf_compass.hdf5_model import HDF5Group, HDF5Store
from hdf_compass.utils import data_url

def test_hdf5_model():
    """Test the HDF5 model with the tall.h5 file"""
    try:
        # Get the URL to the test file
        url = os.path.join(data_url(), "hdf5", "tall.h5")
        log.debug("Testing with file: %s" % url)
        
        # Create a store
        s = store(HDF5Store, url)
        log.debug("Created store: %s" % s)
        
        # Create a container
        c = container(HDF5Store, url, HDF5Group, "/")
        log.debug("Created container: %s" % c)
        
        # List the contents of the container
        log.debug("Container contents:")
        for item in c:
            log.debug("  %s" % item.display_name)
            
        return True
    except Exception as e:
        log.error("Test failed: %s" % str(e))
        return False

if __name__ == "__main__":
    success = test_hdf5_model()
    sys.exit(0 if success else 1) 


##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

"""
Implements a viewer frame for compass_model.Array.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import wx
import wx.stc as stc

log = logging.getLogger(__name__)

from ..frame import BaseFrame


if wx.Platform == '__WXMSW__':
    faces = {'times': 'Times New Roman',
             'mono': 'Courier New',
             'helv': 'Arial',
             'other': 'Comic Sans MS',
             'size': 10,
             'size2': 8,
             }
else:
    faces = {'times': 'Times',
             'mono': 'Courier',
             'helv': 'Helvetica',
             'other': 'new century schoolbook',
             'size': 13,
             'size2': 11,
             }


class TextViewerFrame(wx.Frame):
    """ Frame for displaying text content """

    def __init__(self, parent, text, title="Text Viewer"):
        """ Create a new text viewer frame """
        super(TextViewerFrame, self).__init__(parent, title=title, size=(800, 600))

        # Create text control
        self.text_ctrl = wx.TextCtrl(
            self,
            style=wx.TE_MULTILINE | wx.TE_READONLY | wx.TE_RICH2,
            value=text
        )

        # Set font
        if wx.Platform == '__WXMSW__':
            font = wx.Font(10, wx.FONTFAMILY_TELETYPE, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL)
        else:
            font = wx.Font(12, wx.FONTFAMILY_TELETYPE, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL)
        self.text_ctrl.SetFont(font)

        # Create sizer
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.text_ctrl, 1, wx.EXPAND | wx.ALL, 5)
        self.SetSizer(sizer)

        # Create toolbar
        self.toolbar = self.CreateToolBar()
        self.toolbar.AddTool(wx.ID_COPY, "Copy", wx.ArtProvider.GetBitmap(wx.ART_COPY, wx.ART_TOOLBAR))
        self.toolbar.AddTool(wx.ID_SAVEAS, "Save As", wx.ArtProvider.GetBitmap(wx.ART_FILE_SAVE_AS, wx.ART_TOOLBAR))
        self.toolbar.Realize()

        # Bind events
        self.Bind(wx.EVT_MENU, self.on_copy, id=wx.ID_COPY)
        self.Bind(wx.EVT_MENU, self.on_save_as, id=wx.ID_SAVEAS)

    def on_copy(self, evt):
        """ Handle copy menu item """
        self.text_ctrl.Copy()

    def on_save_as(self, evt):
        """ Handle save as menu item """
        with wx.FileDialog(
            self,
            "Save text file",
            wildcard="Text files (*.txt)|*.txt",
            style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT
        ) as dlg:
            if dlg.ShowModal() == wx.ID_OK:
                with open(dlg.GetPath(), 'w') as f:
                    f.write(self.text_ctrl.GetValue())


class XmlStc(stc.StyledTextCtrl):
    def __init__(self, parent, xml_string, keywords=None):
        stc.StyledTextCtrl.__init__(self, parent, -1)
        self.SetLexer(stc.STC_LEX_XML)
        if keywords is not None:
            self.SetKeyWords(0, keywords)

        # Global default styles for all languages
        self.StyleSetSpec(stc.STC_STYLE_DEFAULT, "face:%(mono)s,size:%(size)d" % faces)
        self.StyleClearAll()  # Reset all to be like the default

        # Global default styles for all languages
        self.StyleSetSpec(stc.STC_STYLE_DEFAULT, "face:%(mono)s,size:%(size)d" % faces)
        self.StyleSetSpec(stc.STC_STYLE_LINENUMBER, "back:#C0C0C0,face:%(helv)s,size:%(size2)d" % faces)
        self.StyleSetSpec(stc.STC_STYLE_CONTROLCHAR, "face:%(other)s" % faces)
        self.StyleSetSpec(stc.STC_STYLE_BRACELIGHT, "fore:#FFFFFF,back:#0000FF,bold")
        self.StyleSetSpec(stc.STC_STYLE_BRACEBAD, "fore:#000000,back:#FF0000,bold")

        self.StyleSetSpec(stc.STC_H_VALUE, "fore:#0000cc,face:%(helv)s,size:%(size)d" % faces)
        self.StyleSetSpec(stc.STC_H_DEFAULT, "fore:#0000cc,face:%(helv)s,size:%(size)d" % faces)
        self.StyleSetSpec(stc.STC_H_ENTITY, "fore:#0000cc,face:%(helv)s,size:%(size)d" % faces)
        # initial XML tag
        self.StyleSetSpec(stc.STC_H_XMLSTART, "fore:#ffcccc,bold,face:%(helv)s,size:%(size)d" % faces)
        self.StyleSetSpec(stc.STC_H_XMLEND, "fore:#ffcccc,bold,face:%(helv)s,size:%(size)d" % faces)
        # XML tags
        self.StyleSetSpec(stc.STC_H_TAG, "fore:#555555,face:%(helv)s,size:%(size)d" % faces)
        self.StyleSetSpec(stc.STC_H_TAGEND, "fore:#555555,face:%(helv)s,size:%(size)d" % faces)
        self.StyleSetSpec(stc.STC_H_TAGUNKNOWN, "fore:#555555,face:%(helv)s,size:%(size)d" % faces)
        # XML attributes
        self.StyleSetSpec(stc.STC_H_ATTRIBUTE, "fore:#BDB76B,face:%(helv)s,size:%(size)d" % faces)
        self.StyleSetSpec(stc.STC_H_ATTRIBUTEUNKNOWN, "fore:#BDB76B,face:%(helv)s,size:%(size)d" % faces)
        # XML comments and quotes
        self.StyleSetSpec(stc.STC_H_COMMENT, "fore:#888888,face:%(helv)s,size:%(size)d" % faces)
        self.StyleSetSpec(stc.STC_H_SINGLESTRING, "fore:#aa2222,face:%(helv)s,size:%(size)d" % faces)
        self.StyleSetSpec(stc.STC_H_DOUBLESTRING, "fore:#aa2222,face:%(helv)s,size:%(size)d" % faces)
        self.StyleSetSpec(stc.STC_H_SGML_SIMPLESTRING, "fore:#aa2222,face:%(helv)s,size:%(size)d" % faces)
        self.StyleSetSpec(stc.STC_H_SGML_DOUBLESTRING, "fore:#aa2222,face:%(helv)s,size:%(size)d" % faces)

        # Keyword
        self.StyleSetSpec(stc.STC_P_WORD, "fore:#990000,size:%(size)d" % faces)

        # Caret color
        self.SetCaretForeground("BLUE")
        # Selection background
        self.SetSelBackground(1, '#66CCFF')

        self.SetSelBackground(True, wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT))
        self.SetSelForeground(True, wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHTTEXT))

        self.SetProperty("fold", "1")  # Enable folding
        self.SetProperty("fold.html", "1")  # Enable folding
        self.SetProperty("tab.timmy.whinge.level", "1")  # Highlight tab/space mixing (shouldn't be any)
        self.SetMargins(3, 3)  # Set left and right margins
        self.SetMarginType(1, stc.STC_MARGIN_NUMBER)  # Set up the numbers in the margin for margin #1
        self.SetMarginWidth(1, 40)  # Reasonable value for, say, 4-5 digits using a mono font (40 pix)
        # self.SetViewWhiteSpace(False)
        self.SetWrapMode(stc.STC_WRAP_WORD)
        self.SetUseVerticalScrollBar(True)
        self.SetUseHorizontalScrollBar(True)

        # Setup a margin to hold fold markers
        self.SetMarginType(2, stc.STC_MARGIN_SYMBOL)
        self.SetMarginMask(2, stc.STC_MASK_FOLDERS)
        self.SetMarginSensitive(2, True)
        self.SetMarginWidth(2, 15)
        # marker style
        self.MarkerDefine(stc.STC_MARKNUM_FOLDEREND, stc.STC_MARK_BOXPLUSCONNECTED, "white", "black")
        self.MarkerDefine(stc.STC_MARKNUM_FOLDEROPENMID, stc.STC_MARK_BOXMINUSCONNECTED, "white", "black")
        self.MarkerDefine(stc.STC_MARKNUM_FOLDERMIDTAIL, stc.STC_MARK_TCORNER, "white", "black")
        self.MarkerDefine(stc.STC_MARKNUM_FOLDERTAIL, stc.STC_MARK_LCORNER, "white", "black")
        self.MarkerDefine(stc.STC_MARKNUM_FOLDERSUB, stc.STC_MARK_VLINE, "white", "black")
        self.MarkerDefine(stc.STC_MARKNUM_FOLDER, stc.STC_MARK_BOXPLUS, "white", "#cccccc")
        self.MarkerDefine(stc.STC_MARKNUM_FOLDEROPEN, stc.STC_MARK_BOXMINUS, "white", "black")
        stc.EVT_STC_MARGINCLICK(self, -1, self.on_margin_click)

        self.SetText(xml_string)
        self.SetEditable(False)

    def on_margin_click(self, evt):
        """ Folds and unfolds as needed """

        if evt.GetMargin() == 2:
            if evt.GetShift() and evt.GetControl():
                self.fold_all()

            else:
                line_clicked = self.LineFromPosition(evt.GetPosition())
                if self.GetFoldLevel(line_clicked) & stc.STC_FOLDLEVELHEADERFLAG:
                    if evt.GetShift():
                        self.SetFoldExpanded(line_clicked, True)
                        self.expand_item(line_clicked, True, True, 1)
                    elif evt.GetControl():
                        if self.GetFoldExpanded(line_clicked):
                            self.SetFoldExpanded(line_clicked, False)
                            self.expand_item(line_clicked, False, True, 0)
                        else:
                            self.SetFoldExpanded(line_clicked, True)
                            self.expand_item(line_clicked, True, True, 100)
                    else:
                        self.ToggleFold(line_clicked)

    def fold_all(self):
        line_count = self.GetLineCount()
        expanding = True

        # find out if we are folding or unfolding
        for line_num in range(line_count):
            if self.GetFoldLevel(line_num) & stc.STC_FOLDLEVELHEADERFLAG:
                expanding = not self.GetFoldExpanded(line_num)
                break

        line_num = 0
        while line_num < line_count:
            level = self.GetFoldLevel(line_num)
            if level & stc.STC_FOLDLEVELHEADERFLAG and (level & stc.STC_FOLDLEVELNUMBERMASK) == stc.STC_FOLDLEVELBASE:

                if expanding:
                    self.SetFoldExpanded(line_num, True)
                    line_num = self.expand_item(line_num, True)
                    line_num -= 1
                else:
                    last_child = self.GetLastChild(line_num, -1)
                    self.SetFoldExpanded(line_num, False)
                    if last_child > line_num:
                        self.HideLines(line_num + 1, last_child)

            line_num += 1

    def expand_item(self, line, do_expand, force=False, vis_levels=0, level=-1):
        last_child = self.GetLastChild(line, level)
        line += 1
        while line <= last_child:
            if force:
                if vis_levels > 0:
                    self.ShowLines(line, line)
                else:
                    self.HideLines(line, line)
            else:
                if do_expand:
                    self.ShowLines(line, line)

            if level == -1:
                level = self.GetFoldLevel(line)

            if level & stc.STC_FOLDLEVELHEADERFLAG:
                if force:
                    if vis_levels > 1:
                        self.SetFoldExpanded(line, True)
                    else:
                        self.SetFoldExpanded(line, False)
                    line = self.expand_item(line, do_expand, force, vis_levels - 1)

                else:
                    if do_expand and self.GetFoldExpanded(line):
                        line = self.expand_item(line=line, do_expand=True, force=force, vis_levels=(vis_levels - 1))
                    else:
                        line = self.expand_item(line=line, do_expand=False, force=force, vis_levels=(vis_levels - 1))
            else:
                line += 1

        return line




##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

"""
Defines the left-hand side information panel used to display context info.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import wx

import logging
log = logging.getLogger(__name__)

from hdf_compass import compass_model
from hdf_compass.utils import is_win


# Info panel width
PANEL_WIDTH = 200

# Size of the main title font
FONTSIZE = 16 if is_win else 18


class InfoPanel(wx.Panel):
    """
    Panel displaying general information about the selected object.

    Designed to be displayed vertically; sets its own width (PANEL_WIDTH).
    """

    def __init__(self, parent):
        """ Constructor.  *parent* should be the containing frame.
        """
        wx.Panel.__init__(self, parent, size=(PANEL_WIDTH, 10), style=wx.BORDER_NONE)

        font = wx.Font(FONTSIZE, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL)

        # Sidebar title text
        self.name_text = wx.StaticText(self, style=wx.ALIGN_LEFT | wx.ST_ELLIPSIZE_MIDDLE, size=(PANEL_WIDTH - 40, 30))
        self.name_text.SetFont(font)

        # Sidebar icon (see display method)
        self.static_bitmap = None

        # Descriptive text below the icon
        self.prop_text = wx.StaticText(self, style=wx.ALIGN_LEFT)

        self.sizer = sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.name_text, 0, wx.LEFT | wx.TOP | wx.RIGHT, border=20)
        sizer.Add(self.prop_text, 0, wx.LEFT | wx.RIGHT, border=20)
        sizer.AddStretchSpacer(1)
        self.SetSizer(sizer)

        self.SetBackgroundColour(wx.Colour(255, 255, 255))

    def display(self, node):
        """ Update displayed information on the node.

        See the get* methods for specifics on what's displayed.
        """

        self.name_text.SetLabel(node.display_name)
        self.prop_text.SetLabel(describe(node))

        if self.static_bitmap is not None:
            self.sizer.Detach(self.static_bitmap)
            self.static_bitmap.Destroy()

        # We load the PNG icon directly from the appropriate Node class
        png = wx.Bitmap(type(node).icons[64], wx.BITMAP_TYPE_ANY)
        self.static_bitmap = wx.StaticBitmap(self, wx.ID_ANY, png)
        self.sizer.Insert(1, self.static_bitmap, 0, wx.ALL, border=20)
        self.sizer.Layout()
        self.Layout()


def describe(node):
    """ Return a (possibly multi-line) text description of a node. """
    desc = "%s\n\n" % type(node).class_kind

    if isinstance(node, compass_model.Array):
        desc += "Shape\n%s\n\nType\n%s\n" % \
                (node.shape, dtype_text(node.dtype))

    elif isinstance(node, compass_model.Container):
        desc += "%d items\n" % len(node)
    
    if not isinstance(node, compass_model.KeyValue):    
        # see if there is a key-value handler for this node
        handlers = node.store.gethandlers(node.key)
        for h in handlers:
            
            kv_node = h(node.store, node.key)
            if isinstance(kv_node, compass_model.KeyValue):
                num_keys = len(kv_node.keys)
                if num_keys > 0:
                    desc += "\n%d %s\n" % (len(kv_node.keys), type(kv_node).class_kind)
            
    return desc


def dtype_text(dt):
    """ String description appropriate for a NumPy dtype """

    if dt is None:
        return "Unknown (dtype not available)"

    try:
        log.debug("dtype kind: %s, size: %d" % (dt.kind, dt.itemsize))

        if dt.names is not None:
            log.debug("dtype names: %s" % ",".join(n for n in dt.names))
            return "Compound (%d fields)" % len(dt.names)
        if dt.kind == 'f':
            return "%d-byte floating point" % dt.itemsize
        if dt.kind == 'u':
            return "%d-byte unsigned integer" % dt.itemsize
        if dt.kind == 'i':
            return "%d-byte signed integer" % dt.itemsize
        if dt.kind == 'S':
            return "ASCII String (%d characters)" % dt.itemsize
        if dt.kind == 'U':
            return "Unicode String (%d characters)" % dt.itemsize
        return "Unknown"
    except (AttributeError, TypeError):
        return "Unknown (dtype not accessible)"



"""
This module provides classes and methods to inject WRP (IoWarp) CTE adapters.
The WrpAdapters interceptor enables interception of various I/O APIs (POSIX,
MPI-IO, STDIO, HDF5 VFD, NVIDIA GDS) and routes them to the Content Transfer
Engine for intelligent data placement and transfer.
"""
from jarvis_cd.core.pkg import Interceptor
from jarvis_cd.util import SizeType
import pathlib
import os
import yaml


class WrpAdapters(Interceptor):
    """
    WRP CTE Adapters Interceptor.

    This interceptor enables I/O interception for various APIs by setting up
    LD_PRELOAD with the appropriate WRP CTE adapter libraries. It supports:
    - POSIX I/O (read, write, open, close, etc.)
    - MPI-IO (MPI_File_* operations)
    - STDIO (fread, fwrite, fopen, etc.)
    - HDF5 VFD (Virtual File Driver for HDF5)
    - NVIDIA GDS (GPUDirect Storage)
    """

    def _init(self):
        """
        Initialize the interceptor.

        This method is called during package instantiation.
        No configuration is available yet at this stage.
        """
        self.cae_config_path = None

    def _configure_menu(self):
        """
        Define configuration options for the WRP adapters interceptor.

        Returns:
            List[Dict]: Configuration parameters for adapter selection.
        """
        return [
            {
                'name': 'posix',
                'msg': 'Intercept POSIX I/O operations',
                'type': bool,
                'default': False,
                'help': 'Intercepts read, write, open, close, lseek, etc.'
            },
            {
                'name': 'mpiio',
                'msg': 'Intercept MPI-IO operations',
                'type': bool,
                'default': False,
                'help': 'Intercepts MPI_File_* operations'
            },
            {
                'name': 'stdio',
                'msg': 'Intercept STDIO operations',
                'type': bool,
                'default': False,
                'help': 'Intercepts fread, fwrite, fopen, fclose, etc.'
            },
            {
                'name': 'vfd',
                'msg': 'Intercept HDF5 I/O via VFD',
                'type': bool,
                'default': False,
                'help': 'Enables HDF5 Virtual File Driver for CTE'
            },
            {
                'name': 'nvidia_gds',
                'msg': 'Intercept NVIDIA GDS I/O',
                'type': bool,
                'default': False,
                'help': 'Intercepts NVIDIA GPUDirect Storage operations'
            },
            {
                'name': 'include',
                'msg': 'Path patterns to include for interception (list)',
                'type': list,
                'default': ['.*test_hermes.*'],
                'help': 'List of regex patterns for paths to intercept. Supports env var expansion ($HOME, ${VAR}, ~). Example: [".*test_hermes.*", "/data/.*", "$HOME/scratch/.*"]'
            },
            {
                'name': 'exclude',
                'msg': 'Path patterns to exclude from interception (list)',
                'type': list,
                'default': [],
                'help': 'List of regex patterns for paths to exclude from interception. Takes precedence based on specificity (longer paths checked first). Example: ["/tmp/.*", "/dev/.*"]'
            },
            {
                'name': 'adapter_page_size',
                'msg': 'Adapter page size for I/O operations',
                'type': str,
                'default': '1M',
                'help': 'Page size for adapter operations. Supports SizeType format (e.g., "4K", "1M", "4096")'
            },
        ]

    def _configure(self, **kwargs):
        """
        Configure the WRP adapters interceptor.

        This method finds the adapter libraries, stores their paths
        in the environment, and generates the CAE configuration file.

        Args:
            **kwargs: Configuration parameters (automatically updated to self.config)

        Raises:
            Exception: If no adapter is selected or if a selected adapter library
                      cannot be found.
        """
        has_one = False

        if self.config['posix']:
            posix_lib = self.find_library('wrp_cte_posix')
            if posix_lib is None:
                raise Exception('Could not find wrp_cte_posix library')
            self.env['WRP_CTE_POSIX'] = posix_lib
            self.env['WRP_CTE_ROOT'] = str(pathlib.Path(posix_lib).parent.parent)
            self.log(f'Found libwrp_cte_posix.so at {posix_lib}')
            has_one = True

        if self.config['mpiio']:
            mpiio_lib = self.find_library('wrp_cte_mpiio')
            if mpiio_lib is None:
                raise Exception('Could not find wrp_cte_mpiio library')
            self.env['WRP_CTE_MPIIO'] = mpiio_lib
            self.env['WRP_CTE_ROOT'] = str(pathlib.Path(mpiio_lib).parent.parent)
            self.log(f'Found libwrp_cte_mpiio.so at {mpiio_lib}')
            has_one = True

        if self.config['stdio']:
            stdio_lib = self.find_library('wrp_cte_stdio')
            if stdio_lib is None:
                raise Exception('Could not find wrp_cte_stdio library')
            self.env['WRP_CTE_STDIO'] = stdio_lib
            self.env['WRP_CTE_ROOT'] = str(pathlib.Path(stdio_lib).parent.parent)
            self.log(f'Found libwrp_cte_stdio.so at {stdio_lib}')
            has_one = True

        if self.config['vfd']:
            vfd_lib = self.find_library('wrp_cte_vfd')
            if vfd_lib is None:
                raise Exception('Could not find wrp_cte_vfd library')
            self.env['WRP_CTE_VFD'] = vfd_lib
            self.env['WRP_CTE_ROOT'] = str(pathlib.Path(vfd_lib).parent.parent)
            self.log(f'Found libwrp_cte_vfd.so at {vfd_lib}')
            has_one = True

        if self.config['nvidia_gds']:
            nvidia_gds_lib = self.find_library('wrp_cte_nvidia_gds')
            if nvidia_gds_lib is None:
                raise Exception('Could not find wrp_cte_nvidia_gds library')
            self.env['WRP_CTE_NVIDIA_GDS'] = nvidia_gds_lib
            self.env['WRP_CTE_ROOT'] = str(pathlib.Path(nvidia_gds_lib).parent.parent)
            self.log(f'Found libwrp_cte_nvidia_gds.so at {nvidia_gds_lib}')
            has_one = True

        if not has_one:
            raise Exception('No WRP CTE adapter selected. Please enable at least one adapter (posix, mpiio, stdio, vfd, or nvidia_gds).')

        # Generate CAE configuration
        self._generate_cae_config()

    def modify_env(self):
        """
        Modify the environment to enable WRP CTE adapter interception.

        This method is called automatically by Jarvis during pipeline start,
        just before the target package's start() method. It modifies the shared
        mod_env to add adapter libraries to LD_PRELOAD.

        The mod_env is shared between the interceptor and the target package,
        so changes made here directly affect the package's execution environment.
        """
        # Set CAE configuration environment variable
        if self.cae_config_path and os.path.exists(self.cae_config_path):
            self.setenv('WRP_CAE_CONF', self.cae_config_path)
            self.log(f"Set WRP_CAE_CONF to {self.cae_config_path}")

        # Add POSIX adapter to LD_PRELOAD
        if self.config['posix']:
            self.prepend_env('LD_PRELOAD', self.env['WRP_CTE_POSIX'])
            self.log(f"Added POSIX adapter to LD_PRELOAD")

        # Add MPI-IO adapter to LD_PRELOAD
        if self.config['mpiio']:
            self.prepend_env('LD_PRELOAD', self.env['WRP_CTE_MPIIO'])
            self.log(f"Added MPI-IO adapter to LD_PRELOAD")

        # Add STDIO adapter to LD_PRELOAD
        if self.config['stdio']:
            self.prepend_env('LD_PRELOAD', self.env['WRP_CTE_STDIO'])
            self.log(f"Added STDIO adapter to LD_PRELOAD")

        # Configure HDF5 VFD (uses plugin path instead of LD_PRELOAD)
        if self.config['vfd']:
            plugin_path_parent = str(pathlib.Path(self.env['WRP_CTE_VFD']).parent)
            self.setenv('HDF5_PLUGIN_PATH', plugin_path_parent)
            self.setenv('HDF5_DRIVER', 'wrp_cte_vfd')
            self.log(f"Configured HDF5 VFD with plugin path: {plugin_path_parent}")

        # Add NVIDIA GDS adapter to LD_PRELOAD
        if self.config['nvidia_gds']:
            self.prepend_env('LD_PRELOAD', self.env['WRP_CTE_NVIDIA_GDS'])
            self.log(f"Added NVIDIA GDS adapter to LD_PRELOAD")

    def _generate_cae_config(self):
        """
        Generate the Content Adapter Engine (CAE) configuration file.

        This creates a YAML configuration file that specifies:
        - include: List of regex patterns for paths to include
        - exclude: List of regex patterns for paths to exclude
        - adapter_page_size: Page size for adapter I/O operations

        Paths are checked in order of specificity (longest path first).
        If a path matches an include pattern, it's intercepted.
        If it matches an exclude pattern, it's not intercepted.
        If no patterns match, the path is excluded by default.

        The configuration file is saved to the shared directory and its path
        is stored in self.cae_config_path for use in modify_env().
        """
        # Parse adapter page size using SizeType
        page_size_bytes = SizeType(self.config['adapter_page_size']).bytes

        # Get include paths from configuration and expand environment variables
        include_patterns = self.config.get('include', ['.*test_hermes.*'])
        expanded_includes = []
        for pattern in include_patterns:
            # Expand environment variables (e.g., $HOME, ${SCRATCH_DIR})
            expanded_pattern = os.path.expandvars(pattern)
            # Also expand user home directory (e.g., ~)
            expanded_pattern = os.path.expanduser(expanded_pattern)
            expanded_includes.append(expanded_pattern)

            # Log if pattern was expanded
            if expanded_pattern != pattern:
                self.log(f"Expanded include pattern: {pattern} -> {expanded_pattern}")

        # Get exclude paths from configuration and expand environment variables
        exclude_patterns = self.config.get('exclude', [])
        expanded_excludes = []
        for pattern in exclude_patterns:
            # Expand environment variables
            expanded_pattern = os.path.expandvars(pattern)
            # Also expand user home directory
            expanded_pattern = os.path.expanduser(expanded_pattern)
            expanded_excludes.append(expanded_pattern)

            # Log if pattern was expanded
            if expanded_pattern != pattern:
                self.log(f"Expanded exclude pattern: {pattern} -> {expanded_pattern}")

        # Build CAE configuration
        cae_config = {
            'include': expanded_includes,
            'exclude': expanded_excludes,
            'adapter_page_size': page_size_bytes
        }

        # Save configuration to shared directory
        self.cae_config_path = os.path.join(self.shared_dir, 'cae_config.yaml')

        try:
            with open(self.cae_config_path, 'w') as f:
                yaml.dump(cae_config, f, default_flow_style=False, indent=2)

            # Set environment variable immediately after creating config file
            self.env['WRP_CAE_CONF'] = self.cae_config_path

            self.log(f"CAE configuration written to: {self.cae_config_path}")
            self.log(f"Set WRP_CAE_CONF={self.cae_config_path}")

            # Log configuration details
            if expanded_includes:
                self.log(f"Include patterns ({len(expanded_includes)}): {', '.join(expanded_includes)}")
            else:
                self.log("No include patterns specified - using default: .*test_hermes.*")

            if expanded_excludes:
                self.log(f"Exclude patterns ({len(expanded_excludes)}): {', '.join(expanded_excludes)}")
            else:
                self.log("No exclude patterns specified")

            self.log(f"Adapter page size: {SizeType(page_size_bytes).to_human_readable()}")
            self.log("Path matching: Most specific pattern wins, default is exclude")

        except Exception as e:
            self.log(f"Error writing CAE configuration: {e}")
            raise

    def clean(self):
        """
        Clean up interceptor data.

        WRP adapters typically don't create persistent data that needs cleanup.
        The CAE configuration file is left in place as it may be useful for
        debugging or reuse.
        """
        pass



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

"""
This class is used to map between HDF5 type representations and numpy types   
 
"""
import numpy as np
from h5py.h5t import special_dtype 
from h5py.h5t import check_dtype
from h5py.h5r import Reference
from h5py.h5r import RegionReference


"""
Convert the given type item  to a predefined type string for 
predefined integer and floating point types ("H5T_STD_I64LE", et. al).
For compound types, recursively iterate through the typeItem and do same
conversion for fields of the compound type.
"""    
def getTypeResponse(typeItem):
    response = None
    if 'uuid' in typeItem:
        # committed type, just return uuid
        response = 'datatypes/' + typeItem['uuid']
    elif typeItem['class'] == 'H5T_INTEGER' or  typeItem['class'] == 'H5T_FLOAT':
        # just return the class and base for pre-defined types
        response = {}
        response['class'] = typeItem['class']
        response['base'] = typeItem['base']
    elif typeItem['class'] == 'H5T_OPAQUE':
        response = {}
        response['class'] = 'H5T_OPAQUE'
        response['size'] = typeItem['size']
    elif typeItem['class'] == 'H5T_REFERENCE':
        response = {}
        response['class'] = 'H5T_REFERENCE'
        response['base'] = typeItem['base']
    elif typeItem['class'] == 'H5T_COMPOUND':
        response = {}
        response['class'] = 'H5T_COMPOUND'
        fieldList = []
        for field in typeItem['fields']:
            fieldItem = { }
            fieldItem['name'] = field['name']
            fieldItem['type'] = getTypeResponse(field['type'])  # recursive call
            fieldList.append(fieldItem)
        response['fields'] = fieldList
    else:
        response = {}   # otherwise, return full type
        for k in typeItem.keys():
            if k == 'base':
                if type(typeItem[k]) == dict:
                    response[k] = getTypeResponse(typeItem[k])  # recursive call
                else:
                    response[k] = typeItem[k]  # predefined type
            elif k not in ('size', 'base_size'):
                response[k] = typeItem[k]
    return response
    
            
"""
    Return type info.
          For primitive types, return string with typename
          For compound types return array of dictionary items
"""
def getTypeItem(dt):
    type_info = {}
    if len(dt) <= 1:
        type_info = getTypeElement(dt)
    else:
        names = dt.names
        type_info['class'] = 'H5T_COMPOUND'
        fields = []
        for name in names:
            field = { 'name': name }
            field['type'] = getTypeElement(dt[name])  
            fields.append(field)
            type_info['fields'] = fields
    return type_info
             
"""
    Get element type info - either a complete type or element of a compound type
    Returns dictionary
    Note: only getTypeItem should call this!
"""
            
def getTypeElement(dt):
    if len(dt) > 1:
        raise Exception("unexpected numpy type passed to getTypeElement")
    
    type_info = {}
         
    if dt.kind == 'O':
        # numpy object type - assume this is a h5py variable length extension
        h5t_check = check_dtype(vlen=dt)
        if h5t_check is not None:
            
            if h5t_check == str:
                type_info['class'] = 'H5T_STRING'
                type_info['length'] = 'H5T_VARIABLE'
                type_info['charSet'] = 'H5T_CSET_ASCII'
                type_info['strPad'] = 'H5T_STR_NULLTERM'
            elif h5t_check == unicode:
                type_info['class'] = 'H5T_STRING'
                type_info['length'] = 'H5T_VARIABLE'
                type_info['charSet'] = 'H5T_CSET_UTF8'
                type_info['strPad'] = 'H5T_STR_NULLTERM'
            elif type(h5t_check) == np.dtype:
                # vlen data
                type_info['class'] = 'H5T_VLEN'
                type_info['size'] = 'H5T_VARIABLE'
                type_info['base'] = getBaseType(h5t_check)  
            else:
                #unknown vlen type
                raise TypeError("Unknown h5py vlen type: " + h5t_check)
        else:
            # check for reference type
            h5t_check = check_dtype(ref=dt)
            if h5t_check is not None:
                type_info['class'] = 'H5T_REFERENCE'
              
                if h5t_check is Reference:
                    type_info['base'] = 'H5T_STD_REF_OBJ'  # objref
                elif h5t_check is RegionReference:
                    type_info['base'] = 'H5T_STD_REF_DSETREG'  # region ref
                else:
                    raise TypeError("unexpected reference type")
            else:     
                raise TypeError("unknown object type")
    elif dt.kind == 'V':
        baseType = getBaseType(dt)
        if dt.shape:
            # array type
            type_info['dims'] = dt.shape
            type_info['class'] = 'H5T_ARRAY'
            type_info['base'] = baseType
        elif baseType['class'] == 'H5T_OPAQUE':
            # expecting this to be an opaque type
            type_info = baseType  # just promote the base type
        else:
            raise TypeError("unexpected Void type")
    elif dt.kind == 'S':
        # String type
        baseType = getBaseType(dt)
        type_info = baseType  # just use base type
    elif dt.kind == 'U':
        # Unicode String type
        baseType = getBaseType(dt)
        type_info = baseType  # just use base type
    elif dt.kind == 'i' or dt.kind == 'u':
        # integer type
        baseType = getBaseType(dt)
        # numpy integer type - but check to see if this is the hypy 
        # enum extension
        mapping = check_dtype(enum=dt)  
            
        if mapping:
            # yes, this is an enum!
            type_info['class'] = 'H5T_ENUM'
            type_info['mapping'] = mapping
            type_info['base'] = baseType
        else:
            type_info = baseType  # just use base type
    
    elif dt.kind == 'f':
        # floating point type
        baseType = getBaseType(dt)
        type_info = baseType  # just use base type
    else:
        # unexpected kind
        raise TypeError("unexpected dtype kind: " + dt.kind)
        
    return type_info
        
"""
Get Base type info for given type element.
"""    
def getBaseType(dt):
    if len(dt) > 1:
        raise TypeError("unexpected numpy type passed to getTypeElement")
             
    predefined_int_types = {
        'int8':    'H5T_STD_I8',
        'uint8':   'H5T_STD_U8',
        'int16':   'H5T_STD_I16',
        'uint16':  'H5T_STD_U16',
        'int32':   'H5T_STD_I32',
        'uint32':  'H5T_STD_U32',
        'int64':   'H5T_STD_I64',
        'uint64':  'H5T_STD_U64'
    }
    predefined_float_types = {
        'float32': 'H5T_IEEE_F32',
        'float64': 'H5T_IEEE_F64'
    }
    type_info = {}
    
    #type_info['base_size'] = dt.base.itemsize
         
    # primitive type
    if dt.base.kind == 'S':
        # Fixed length string type
        type_info['class'] = 'H5T_STRING' 
        type_info['charSet'] = 'H5T_CSET_ASCII'
        type_info['length'] = dt.base.itemsize
        type_info['strPad'] = 'H5T_STR_NULLPAD'    
    elif dt.base.kind == 'V':
            type_info['class'] = 'H5T_OPAQUE'
            type_info['size'] = dt.itemsize
            type_info['tag'] = ''  # todo - determine tag
    elif dt.base.kind == 'i' or dt.base.kind == 'u':    
        type_info['class'] = 'H5T_INTEGER'
        byteorder = 'LE'
        if dt.base.byteorder == '>':
            byteorder = 'BE'
        if dt.base.name in predefined_int_types:
            #maps to one of the HDF5 predefined types
            type_info['base'] = predefined_int_types[dt.base.name] + byteorder  
    elif dt.base.kind == 'f':
        type_info['class'] = 'H5T_FLOAT'
        byteorder = 'LE'
        if dt.base.byteorder == '>':
            byteorder = 'BE'
        if dt.base.name in predefined_float_types:
            #maps to one of the HDF5 predefined types
            type_info['base'] = predefined_float_types[dt.base.name] + byteorder 
    elif dt.base.kind == 'O':
        # check for reference type
        h5t_check = check_dtype(ref=dt)
        if h5t_check is not None:
            type_info['class'] = 'H5T_REFERENCE' 
            if h5t_check is Reference:
                type_info['base'] = 'H5T_STD_REF_OBJ'  # objref
            elif h5t_check is RegionReference:
                type_info['base'] = 'H5T_STD_REF_DSETREG'  # region ref
            else:
                raise TypeError("unexpected reference type")
        else:     
            raise TypeError("unknown object type")
    else:
        # unexpected kind
        raise TypeError("unexpected dtype base kind: " + dt.base.kind)
    
    return type_info
 

def getNumpyTypename(hdf5TypeName, typeClass=None):
    predefined_int_types = {
          'H5T_STD_I8':  'i1', 
          'H5T_STD_U8':  'u1',
          'H5T_STD_I16': 'i2', 
          'H5T_STD_U16': 'u2',
          'H5T_STD_I32': 'i4', 
          'H5T_STD_U32': 'u4',
          'H5T_STD_I64': 'i8',
          'H5T_STD_U64': 'u8' 
    }
    predefined_float_types = {
          'H5T_IEEE_F32': 'f4',
          'H5T_IEEE_F64': 'f8'
    }
    
    if len(hdf5TypeName) < 3:
        raise Exception("Type Error: invalid typename: ")
    endian = '<'  # default endian
    key = hdf5TypeName
    if hdf5TypeName.endswith('LE'):
        key = hdf5TypeName[:-2]
    elif hdf5TypeName.endswith('BE'):
        key = hdf5TypeName[:-2]
        endian = '>'
        
    if key in predefined_int_types and (typeClass == None or 
            typeClass == 'H5T_INTEGER'):
        return endian + predefined_int_types[key]
    if key in predefined_float_types and (typeClass == None or 
            typeClass == 'H5T_FLOAT'):
        return endian + predefined_float_types[key]
    raise TypeError("Type Error: invalid type")
    
    
def createBaseDataType(typeItem):

    dtRet = None
    if type(typeItem) == str or type(typeItem) == unicode:
        # should be one of the predefined types
        dtName = getNumpyTypename(typeItem)
        dtRet = np.dtype(dtName)
        return dtRet  # return predefined type
        
    if type(typeItem) != dict:
        raise TypeError("Type Error: invalid type")
        
        
    if 'class' not in typeItem:
        raise KeyError("'class' not provided")
    typeClass = typeItem['class']
    
    dims = ''
    if 'dims' in typeItem:  
        dims = None      
        if type(typeItem['dims']) == int:
            dims = (typeItem['dims'])  # make into a tuple
        elif type(typeItem['dims']) not in (list, tuple):
            raise TypeError("expected list or integer for dims")
        else:
            dims = typeItem['dims']
        dims = str(tuple(dims))
        
    if typeClass == 'H5T_INTEGER':
        if 'base' not in typeItem:
            raise KeyError("'base' not provided")      
        baseType = getNumpyTypename(typeItem['base'], typeClass='H5T_INTEGER')
        dtRet = np.dtype(dims + baseType)
    elif typeClass == 'H5T_FLOAT':
        if 'base' not in typeItem:
            raise KeyError("'base' not provided")
        baseType = getNumpyTypename(typeItem['base'], typeClass='H5T_FLOAT')
        dtRet = np.dtype(dims + baseType)
    elif typeClass == 'H5T_STRING':
        if 'length' not in typeItem:
            raise KeyError("'length' not provided")
        if 'charSet' not in typeItem:
            raise KeyError("'charSet' not provided")          
            
        if typeItem['length'] == 'H5T_VARIABLE':
            if dims:
                raise TypeError("ArrayType is not supported for variable len types")
            if typeItem['charSet'] == 'H5T_CSET_ASCII':
                dtRet = special_dtype(vlen=str)
            elif typeItem['charSet'] == 'H5T_CSET_UTF8':
                dtRet = special_dtype(vlen=unicode)
            else:
                raise TypeError("unexpected 'charSet' value")
        else:
            nStrSize = typeItem['length']
            if type(nStrSize) != int:
                raise TypeError("expecting integer value for 'length'")
            type_code = None
            if typeItem['charSet'] == 'H5T_CSET_ASCII':
                type_code = 'S'
            elif typeItem['charSet'] == 'H5T_CSET_UTF8':
                raise TypeError("fixed-width unicode strings are not supported")
            else:
                raise TypeError("unexpected 'charSet' value")
            dtRet = np.dtype(dims + type_code + str(nStrSize))  # fixed size string
    elif typeClass == 'H5T_VLEN':
        if dims:
            raise TypeError("ArrayType is not supported for variable len types")
        if 'base' not in typeItem:
            raise KeyError("'base' not provided") 
        baseType = createBaseDataType(typeItem['base'])
        dtRet = special_dtype(vlen=np.dtype(baseType))
    elif typeClass == 'H5T_OPAQUE':
        if dims:
            raise TypeError("Opaque Type is not supported for variable len types")
        if 'size' not in typeItem:
            raise KeyError("'size' not provided")
        nSize = int(typeItem['size'])
        if nSize <= 0:
            raise TypeError("'size' must be non-negative")
        dtRet = np.dtype('V' + str(nSize))
    elif typeClass == 'H5T_ARRAY':
        if not dims:
            raise KeyError("'dims' must be provided for array types")
        if 'base' not in typeItem:
            raise KeyError("'base' not provided") 
        arrayBaseType = typeItem['base']
        if type(arrayBaseType) is dict:
            if "class" not in arrayBaseType:
                raise KeyError("'class' not provided for array base type")
            if arrayBaseType["class"] not in ('H5T_INTEGER', 'H5T_FLOAT', 'H5T_STRING'):
                raise TypeError("Array Type base type must be integer, float, or string")
                
        baseType = createDataType(arrayBaseType)    
        dtRet = np.dtype(dims+baseType.str)
        return dtRet  # return predefined type    
    elif typeClass == 'H5T_REFERENCE':
        if 'base' not in typeItem:
            raise KeyError("'base' not provided") 
        if typeItem['base'] == 'H5T_STD_REF_OBJ':
        	dtRet = special_dtype(ref=Reference)
        elif typeItem['base'] == 'H5T_STD_REF_DSETREG':
        	dtRet = special_dtype(ref=RegionReference)
        else:
            raise TypeError("Invalid base type for reference type")
        
    else:
        raise TypeError("Invalid type class")
        
      
    return dtRet  
    
def createDataType(typeItem):
    dtRet = None
    if type(typeItem) == str or type(typeItem) == unicode:
        # should be one of the predefined types
        dtName = getNumpyTypename(typeItem)
        dtRet = np.dtype(dtName)
        return dtRet  # return predefined type
        
    if type(typeItem) != dict:
        raise TypeError("invalid type")
        
        
    if 'class' not in typeItem:
        raise KeyError("'class' not provided")
    typeClass = typeItem['class']
    
    if typeClass == 'H5T_COMPOUND':
        if 'fields' not in typeItem:
            raise KeyError("'fields' not provided for compound type")
        fields = typeItem['fields']
        if type(fields) is not list:
            raise TypeError("Type Error: expected list type for 'fields'")
        if not fields:
            raise KeyError("no 'field' elements provided")
        subtypes = []
        for field in fields:
             
            if type(field) != dict:
                raise TypeError("Expected dictionary type for field")
            if 'name' not in field:
                raise KeyError("'name' missing from field")
            if 'type' not in field:
                raise KeyError("'type' missing from field")
            field_name = field['name']
            if type(field_name) == unicode:
                # convert to ascii
                ascii_name = field_name.encode('ascii')
                if ascii_name != field_name:
                    raise TypeError("non-ascii field name not allowed")
                field['name'] = ascii_name
                
            dt = createDataType(field['type'])  # recursive call
            if dt is None:
                raise Exception("unexpected error")
            subtypes.append((field['name'], dt))  # append tuple
        dtRet = np.dtype(subtypes)
    else:
        dtRet = createBaseDataType(typeItem)  # create non-compound dt
    return dtRet
        
                
    
        
            
   
